#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Strict ASN.1 DER format for ECDSA signature representation.

The original Bitcoin implementation used OpenSSL to verify
ECDSA signatures in ASN.1 DER representation.
However, OpenSSL does not do strict validation
(e.g. extra padding is ignored) and this changes the transaction
hash value, leading to transaction malleability.
This was fixed by BIP66, activated on block 363,724.

source:
https://github.com/bitcoin/bips/blob/master/bip-0066.mediawiki

BIP66 mandates a strict DER format:

Format:
[0x30][length][0x02][R-length][R][0x02][S-length][S][sighash]

* 0x30 header byte to indicate compound structure
* length: 1-byte length descriptor of the following data,
    excluding the sighash byte
* 0x02 header byte indicating an integer
* r-length: 1-byte length descriptor of the r value that follows
* r: arbitrary-length big-endian r value.
    It must use the shortest possible encoding for
    a positive integers (which means no null bytes at the start,
    except a single one when the next byte has its highest bit set
    to avoid being interpreted as a negative number)
* 0x02 header byte indicating an integer
* s-length: 1-byte length descriptor of the s value that follows
* s: arbitrary-length big-endian s value. Same rules as for r apply
* sighash: 1-byte value indicating what data is hashed
    (not part of the DER signature)

There are 7 bytes of meta-data:

* compound header, compound length,
* value header, r value length,
* value header, s value length
* sighash type (optional)

The ECDSA signature (r, s) should be 64 bytes,
r and s being 32 bytes integers each;
however, integers in DER are signed,
so if the value being encoded is greater than 2^128,
a 33rd byte is added in front.
Bitcoin has a "low s" rule for the s value to be below ec.n,
but it is only a standardness rule miners are allowed to ignore.
Moreover, no such rule exists for r.
"""

from typing import Optional, Tuple, Union

from .alias import Octets
from .curve import Curve
from .curves import secp256k1
from .script import SIGHASHES
from .utils import bytes_from_octets

# (r, s, sighash)
# r and s are the components of a DSASigTuple
DERSigTuple = Tuple[int, int, Optional[bytes]]
# DERSigTuple or DER serialization (bytes or hex-string, with sighash)
DERSig = Union[DERSigTuple, Octets]


def _validate_sig(
    r: int, s: int, sighash: Optional[Octets] = None, ec: Curve = secp256k1
) -> None:
    # check that the DSA/DER signature is correct

    # Fail if r is not [1, n-1]
    if not 0 < r < ec.n:
        raise ValueError(f"r ({hex(r)}) not in [1, n-1]")

    # Fail if s is not [1, n-1]
    if not 0 < s < ec.n:
        raise ValueError(f"s ({hex(s)}) not in [1, n-1]")

    if sighash is not None and sighash not in SIGHASHES:
        m = f"Invalid sighash ({sighash!r})"
        raise ValueError(m)


def _deserialize(sig: DERSig, ec: Curve = secp256k1) -> DERSigTuple:
    """Deserialize a strict ASN.1 DER representation of an ECDSA signature.

    Return r, s, sighash; sighash is None if not available.
    """

    if isinstance(sig, tuple):
        r, s, sighash = sig
    else:
        sig = bytes_from_octets(sig)

        # 73 bytes for secp256k1 (including sighash)
        maxsize = (ec.nsize + 1) * 2 + 6 + 1
        # 1 byte for r, 1 bytes for s, excluding sighash
        minsize = 1 * 2 + 6
        sigsize = len(sig)
        if not minsize <= sigsize <= maxsize:
            errmsg = f"DER signature size ({sigsize}) must be in "
            errmsg += f"{minsize}, {maxsize}]"
            raise ValueError(errmsg)

        if sig[0] != 0x30:
            raise ValueError(
                f"DER signature type must be 0x30 (compound), " "not {hex(sig[0])}"
            )

        # sigsize checks
        leftover = sigsize - 2 - sig[1]
        if leftover == 0:  # no sighash value
            sighash = None
        elif leftover == 1:  # sighash value
            sighash = sig[sigsize - 1 :]
        else:
            msg = f"Declared length ({sig[1]}) does not "
            msg += f"match with actual signature size ({sigsize}) +2 or +3"
            raise ValueError(msg)

        sizeR = sig[3]  # size of the r scalar
        if sizeR == 0:
            raise ValueError("Zero-size integer is not allowed for r")

        if 5 + sizeR >= sigsize:
            raise ValueError("Size of the s scalar must be inside the signature")

        sizeS = sig[5 + sizeR]  # size of the s scalar
        if sizeS == 0:
            raise ValueError("Zero-size integer is not allowed for s")

        if sigsize - sizeR - sizeS != 6 + leftover:
            raise ValueError("Signature size does not match with size of scalars")

        # scalar r
        if sig[2] != 0x02:
            raise ValueError("r scalar must be an integer")

        if sig[4] & 0x80:
            raise ValueError("Negative number is not allowed for r")

        # Null bytes at the start of a scalar are not allowed, unless the
        # scalar would otherwise be interpreted as a negative number
        if sizeR > 1 and sig[4] == 0x00 and not (sig[5] & 0x80):
            raise ValueError("Invalid null bytes at the start of r")

        r = int.from_bytes(sig[4 : 4 + sizeR], byteorder="big")

        # scalar s (offset=2+sizeR with respect to r)
        if sig[sizeR + 4] != 0x02:
            raise ValueError("s scalar must be an integer")

        if sig[sizeR + 6] & 0x80:
            raise ValueError("Negative number is not allowed for s")

        # Null bytes at the start of a scalar are not allowed, unless the
        # scalar would otherwise be interpreted as a negative number
        if sizeS > 1 and sig[sizeR + 6] == 0x00 and not (sig[sizeR + 7] & 0x80):
            raise ValueError("Invalid null bytes at the start of s")

        s = int.from_bytes(sig[6 + sizeR : 6 + sizeR + sizeS], byteorder="big")

    _validate_sig(r, s, sighash, ec)
    return r, s, sighash


def _bytes_from_scalar(scalar: int) -> bytes:
    # scalar is assumed to be in [1, n-1]
    elen = scalar.bit_length()
    esize = elen // 8 + 1  # not a bug: 'highest bit set' padding included here
    n_bytes = scalar.to_bytes(esize, byteorder="big")
    return n_bytes


def _serialize_scalar(scalar: int) -> bytes:
    # scalar is assumed to be in [1, n-1]
    x = _bytes_from_scalar(scalar)
    xsize = len(x).to_bytes(1, byteorder="big")
    return b"\x02" + xsize + x


def _serialize(
    r: int, s: int, sighash: Optional[Octets] = None, ec: Curve = secp256k1
) -> bytes:
    """Serialize an ECDSA signature to strict ASN.1 DER representation.

    Trailing sighash is added if provided.
    """

    # check that it is a valid signature for the given Curve
    _validate_sig(r, s, sighash, ec)
    result = _serialize_scalar(r)
    result += _serialize_scalar(s)
    result = b"\x30" + len(result).to_bytes(1, byteorder="big") + result
    if sighash is None:
        return result

    sighash = bytes_from_octets(sighash, 1)
    return result + sighash



#!/usr/bin/env python3

# Copyright (c) 2017 Pieter Wuille
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

# Copyright (C) 2019-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.


"""Tests for `btclib.bech32address` module.

Some of these tests are originally from
https://github.com/sipa/bech32/tree/master/ref/python,
with the following modifications:

- splitted the original tests.py file in test_bech32.py
  and test_bech32address.py
- checked for raised exceptions instead of assertIsNone
"""

import unittest

import pytest

from btclib.base58address import p2wpkh_p2sh, p2wsh_p2sh
from btclib.bech32address import (
    b32address_from_witness,
    p2wpkh,
    p2wsh,
    witness_from_b32address,
)
from btclib.script import encode
from btclib.secpoint import bytes_from_point, point_from_octets
from btclib.utils import hash160, sha256


class TestSegwitAddress(unittest.TestCase):
    """Unit test class for SegWit addressess."""

    def test_valid_address(self):
        """Test whether valid addresses decode to the correct output"""

        VALID_BC_ADDRESS = [
            [
                "BC1QW508D6QEJXTDG4Y5R3ZARVARY0C5XW7KV8F3T4",
                "0014751e76e8199196d454941c45d1b3a323f1433bd6",
            ],
            [
                (
                    "bc1pw508d6qejxtdg4y5r3zarvary0c5xw7kw508d6qejxtdg4y5r3zarva"
                    "ry0c5xw7k7grplx"
                ),
                (
                    "5128751e76e8199196d454941c45d1b3a323f1433bd6751e76e8199196d"
                    "454941c45d1b3a323f1433bd6"
                ),
            ],
            ["BC1SW50QA3JX3S", "6002751e"],
            [
                "bc1zw508d6qejxtdg4y5r3zarvaryvg6kdaj",
                "5210751e76e8199196d454941c45d1b3a323",
            ],
            [
                " bc1zw508d6qejxtdg4y5r3zarvaryvg6kdaj",  # extra leading space
                "5210751e76e8199196d454941c45d1b3a323",
            ],
        ]
        VALID_TB_ADDRESS = [
            [
                ("tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0s" "l5k7"),
                (
                    "00201863143c14c5166804bd19203356da136c985678cd4d27a1b8c6329"
                    "604903262"
                ),
            ],
            [
                ("tb1qqqqqp399et2xygdj5xreqhjjvcmzhxw4aywxecjdzew6hylgvsesrx" "h6hy"),
                (
                    "0020000000c4a5cad46221b2a187905e5266362b99d5e91c6ce24d165da"
                    "b93e86433"
                ),
            ],
        ]

        for a, hexscript in VALID_BC_ADDRESS + VALID_TB_ADDRESS:
            witvers, witprog, network, _ = witness_from_b32address(a)
            script_pubkey = [witvers, witprog]
            self.assertEqual(encode(script_pubkey).hex(), hexscript)
            address = b32address_from_witness(witvers, witprog, network)
            self.assertEqual(a.lower().strip(), address.decode("ascii"))

    def test_invalid_address(self):
        """Test whether invalid addresses fail to decode"""

        INVALID_ADDRESS = [
            # Invalid human-readable part
            "tc1qw508d6qejxtdg4y5r3zarvary0c5xw7kg3g4ty",
            # Invalid checksum
            "bc1qw508d6qejxtdg4y5r3zarvary0c5xw7kv8f3t5",
            # Invalid witness version
            "BC13W508D6QEJXTDG4Y5R3ZARVARY0C5XW7KN40WF2",
            # Invalid program length
            "bc1rw5uspcuh",
            # Invalid program length
            (
                "bc10w508d6qejxtdg4y5r3zarvary0c5xw7kw508d6qe"
                "jxtdg4y5r3zarvary0c5xw7kw5rljs90"
            ),
            # Invalid program length for witness version 0 (per BIP141)
            "BC1QR508D6QEJXTDG4Y5R3ZARVARYV98GJ9P",
            # Mixed case
            "tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sL5k7",
            # zero padding of more than 4 bits
            "bc1zw508d6qejxtdg4y5r3zarvaryvqyzf3du",
            # Non-zero padding in 8-to-5 conversion
            "tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3pjxtptv",
            # Empty data section
            "bc1gmk9yu",
            # 92 chars
            (
                "bc1qqvpsxqcrqvpsxqcrqvpsxqcrqvpsxqcrqvpsxqcrqv"
                "psxqcrqvpsxqcrqvpsxqcrqvpsxqcrqvpsxqcrqv0jstn5"
            ),
        ]

        for a in INVALID_ADDRESS:
            self.assertRaises(ValueError, witness_from_b32address, a)

    def test_invalid_address_enc(self):
        """Test whether address encoding fails on invalid input"""

        INVALID_ADDRESS_ENC = [
            ("MAINNET", 0, 20),
            ("mainnet", 0, 21),
            ("mainnet", 17, 32),
            ("mainnet", 1, 1),
            ("mainnet", 16, 41),
        ]

        network, version, length = INVALID_ADDRESS_ENC[0]
        self.assertRaises(
            KeyError, b32address_from_witness, version, [0] * length, network
        )

        for network, version, length in INVALID_ADDRESS_ENC[1:]:
            self.assertRaises(
                ValueError, b32address_from_witness, version, [0] * length, network
            )

    def test_b32address_from_witness(self):

        # self-consistency
        addr = b"bc1qg9stkxrszkdqsuj92lm4c7akvk36zvhqw7p6ck"
        wv, wp, network, _ = witness_from_b32address(addr)
        self.assertEqual(b32address_from_witness(wv, wp, network), addr)

        # invalid value -1
        wp = [i for i in wp]  # convert to List[int]
        wp[-1] = -1  # alter the last element with an invalid value
        self.assertRaises(ValueError, b32address_from_witness, wv, wp, network)
        # b32address_from_witness(wv, wp, network)

        # string input
        addr = "bc1qg9stkxrszkdqsuj92lm4c7akvk36zvhqw7p6ck"
        wv, wp, network, _ = witness_from_b32address(addr)
        self.assertEqual(b32address_from_witness(wv, wp, network), addr.encode())

    def test_p2wpkh_p2sh(self):
        # https://matthewdowney.github.io/create-segwit-address.html
        pub = " 03a1af804ac108a8a51782198c2d034b28bf90c8803f5a53f76" "276fa69a4eae77f"

        address = p2wpkh_p2sh(pub)
        self.assertEqual(address, b"36NvZTcMsMowbt78wPzJaHHWaNiyR73Y4g")

        address = p2wpkh_p2sh(pub, "testnet")
        self.assertEqual(address, b"2Mww8dCYPUpKHofjgcXcBCEGmniw9CoaiD2")

        # http://bitcoinscri.pt/pages/segwit_p2sh_p2wpkh
        pub = "02 f118cc409775419a931c57664d0c19c405e856ac0ee2f0e2a41" "37d8250531128"

        address = p2wpkh_p2sh(pub)
        self.assertEqual(address, b"3Mwz6cg8Fz81B7ukexK8u8EVAW2yymgWNd")

    def test_p2wpkh(self):

        # https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki
        # leading/trailing spaces should be tolerated
        pub = " 0279BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D959F2" "815B16F81798"
        addr = b"bc1qw508d6qejxtdg4y5r3zarvary0c5xw7kv8f3t4"
        self.assertEqual(addr, p2wpkh(pub))
        addr = b"tb1qw508d6qejxtdg4y5r3zarvary0c5xw7kxpjzsx"
        self.assertEqual(addr, p2wpkh(pub, "testnet"))

        # http://bitcoinscri.pt/pages/segwit_native_p2wpkh
        pub = "02 530c548d402670b13ad8887ff99c294e67fc18097d236d57880c69" "261b42def7"
        addr = b"bc1qg9stkxrszkdqsuj92lm4c7akvk36zvhqw7p6ck"
        self.assertEqual(addr, p2wpkh(pub))

        _, wp, _, _ = witness_from_b32address(addr)
        self.assertEqual(bytes(wp), hash160(pub))

        # Wrong size (65-bytes) for compressed SEC key
        uncompr_pub = bytes_from_point(point_from_octets(pub), compressed=False)
        self.assertRaises(ValueError, p2wpkh, uncompr_pub)
        # p2wpkh(uncompr_pub)

        # Wrong pubkey size: 34 instead of 33
        self.assertRaises(ValueError, p2wpkh, pub + "00")
        # p2wpkh(pub + '00')

        # Witness program length (21) is not 20
        self.assertRaises(
            ValueError, b32address_from_witness, 0, hash160(pub) + b"\x00"
        )
        # b32address_from_witness(0, hash160(pub) + b'\x00')

    def test_hash_from_bech32(self):
        network = "testnet"
        wv = 0
        wp = 20 * b"\x05"
        addr = b32address_from_witness(wv, wp, network)
        _, wp2, n2, _ = witness_from_b32address(addr)
        self.assertEqual(n2, network)
        self.assertEqual(wp2, wp)

        # witness program length (21) is not 20 or 32
        addr = "tb1qq5zs2pg9q5zs2pg9q5zs2pg9q5zs2pg9q5mpvsef"
        self.assertRaises(ValueError, witness_from_b32address, addr)
        # witness_from_b32address(addr)

    def test_p2wsh_p2sh(self):

        # leading/trailing spaces should be tolerated
        pub = " 0279BE667EF9DCBBAC55A06295CE870B07029BFCDB2D" "CE28D959F2815B16F81798"
        witness_script = [pub, "OP_CHECKSIG"]
        witness_script_bytes = encode(witness_script)
        p2wsh_p2sh(witness_script_bytes)
        p2wsh_p2sh(witness_script_bytes, "testnet")


def test_p2wsh():

    # https://github.com/bitcoin/bips/blob/master/bip-0173.mediawiki
    pub = "02 79BE667EF9DCBBAC55A06295CE870B07029BFCDB2DCE28D" "959F2815B16F81798"
    witness_script = [pub, "OP_CHECKSIG"]
    witness_script_bytes = encode(witness_script)

    addr = b"tb1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3q0sl5k7"
    assert addr == p2wsh(witness_script_bytes, "testnet")
    _, wp, _, _ = witness_from_b32address(addr)
    assert bytes(wp) == sha256(witness_script_bytes)

    addr = b"bc1qrp33g0q5c5txsp9arysrx4k6zdkfs4nce4xj0gdcccefvpysxf3qccfmv3"
    assert addr == p2wsh(witness_script_bytes)
    _, wp, _, _ = witness_from_b32address(addr)
    assert bytes(wp) == sha256(witness_script_bytes)

    assert witness_from_b32address(addr)[1] == sha256(witness_script_bytes)

    errMsg = r"witness program length \(35\) is not 20 or 32"
    with pytest.raises(ValueError, match=errMsg):
        b32address_from_witness(0, witness_script_bytes)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib import bip32
from btclib.base58 import b58encode
from btclib.base58wif import wif_from_prvkey
from btclib.curves import secp256k1 as ec
from btclib.to_prvkey import int_from_prvkey, prvkeyinfo_from_prvkey


class TestToPrvKey(unittest.TestCase):
    def test_int_from_prvkey(self):

        # BIP32
        xprv = (
            "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiC"
            "hkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
        ).encode()
        xprv_str = xprv.decode("ascii")
        xprv_dict = bip32.deserialize(xprv)
        # WIF
        wif = wif_from_prvkey(xprv)
        wif_str = wif.decode("ascii")
        # bytes
        qbytes = xprv_dict["key"][1:]
        qhex = qbytes.hex()
        # int
        q = xprv_dict["q"]
        self.assertEqual(q, int_from_prvkey(q))
        # bytes
        self.assertEqual(q, int_from_prvkey(qbytes))
        self.assertRaises(ValueError, int_from_prvkey, b"\x00" + qbytes)
        self.assertEqual(q, int_from_prvkey(qhex))
        self.assertEqual(q, int_from_prvkey(" " + qhex + " "))
        self.assertRaises(ValueError, int_from_prvkey, qhex + "00")
        # WIF
        self.assertEqual(q, int_from_prvkey(wif))
        self.assertRaises(ValueError, int_from_prvkey, wif + b"\x00")
        self.assertEqual(q, int_from_prvkey(wif_str))
        self.assertEqual(q, int_from_prvkey(" " + wif_str + " "))
        # BIP32
        self.assertEqual(q, int_from_prvkey(xprv))
        self.assertRaises(ValueError, int_from_prvkey, xprv + b"\x00")
        self.assertEqual(q, int_from_prvkey(xprv_str))
        self.assertEqual(q, int_from_prvkey(" " + xprv_str + " "))
        self.assertEqual(q, int_from_prvkey(xprv_dict))

        # wrong private key int
        q = ec.n
        self.assertRaises(ValueError, int_from_prvkey, q)
        # bytes
        qbytes = q.to_bytes(32, byteorder="big")
        qhex = qbytes.hex()
        self.assertRaises(ValueError, int_from_prvkey, qbytes)
        self.assertRaises(ValueError, int_from_prvkey, qhex)
        # WIF
        t = b"\x80" + qbytes + b"\x01"
        wif = b58encode(t)
        wif_str = wif.decode("ascii")
        self.assertRaises(ValueError, int_from_prvkey, wif)
        self.assertRaises(ValueError, int_from_prvkey, wif_str)
        # BIP32
        t = xprv_dict["version"]
        t += xprv_dict["depth"].to_bytes(1, "big")
        t += xprv_dict["parent_fingerprint"]
        t += xprv_dict["index"]
        t += xprv_dict["chain_code"]
        t += b"\x00" + qbytes
        xprv = b58encode(t, 78)
        xprv_str = xprv.decode("ascii")
        self.assertRaises(ValueError, int_from_prvkey, xprv)
        self.assertRaises(ValueError, int_from_prvkey, xprv_str)

        # wrong private key int
        q = 0
        self.assertRaises(ValueError, int_from_prvkey, q)
        # bytes
        qbytes = q.to_bytes(32, byteorder="big")
        qhex = qbytes.hex()
        self.assertRaises(ValueError, int_from_prvkey, qbytes)
        self.assertRaises(ValueError, int_from_prvkey, qhex)
        # WIF
        t = b"\x80" + qbytes + b"\x01"
        wif = b58encode(t)
        wif_str = wif.decode("ascii")
        self.assertRaises(ValueError, int_from_prvkey, wif)
        self.assertRaises(ValueError, int_from_prvkey, wif_str)
        # BIP32
        t = xprv_dict["version"]
        t += xprv_dict["depth"].to_bytes(1, "big")
        t += xprv_dict["parent_fingerprint"]
        t += xprv_dict["index"]
        t += xprv_dict["chain_code"]
        t += b"\x00" + qbytes
        xprv = b58encode(t, 78)
        xprv_str = xprv.decode("ascii")
        self.assertRaises(ValueError, int_from_prvkey, xprv)
        self.assertRaises(ValueError, int_from_prvkey, xprv_str)

        # pub key
        xpub = (
            "xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2g"
            "Z29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8"
        ).encode()
        self.assertRaises(ValueError, int_from_prvkey, xpub)

    def test_info_from_prvkey(self):

        xprv = (
            "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiC"
            "hkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
        ).encode()
        xprv_str = xprv.decode("ascii")
        xprv_dict = bip32.deserialize(xprv)
        wif = wif_from_prvkey(xprv)
        wif_str = wif.decode("ascii")
        ref_tuple = (xprv_dict["q"], "mainnet", True)

        # BIP32
        self.assertEqual(ref_tuple, prvkeyinfo_from_prvkey(xprv, "mainnet"))
        self.assertEqual(ref_tuple, prvkeyinfo_from_prvkey(xprv))
        self.assertEqual(ref_tuple, prvkeyinfo_from_prvkey(xprv_str))
        self.assertEqual(ref_tuple, prvkeyinfo_from_prvkey(" " + xprv_str + " "))
        self.assertEqual(ref_tuple, prvkeyinfo_from_prvkey(xprv_dict))

        # Invalid decoded size: 6 bytes instead of 82
        xpub = "notakey"
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, xpub)
        # prvkeyinfo_from_prvkey(xpub)

        # xkey is not a private one
        xpub = (
            "xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2g"
            "Z29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8"
        ).encode()
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, xpub)
        # prvkeyinfo_from_prvkey(xpub)

        # xkey is not a private one
        xpub_dict = bip32.deserialize(xpub)
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, xpub_dict)
        # prvkeyinfo_from_prvkey(xpub_dict)

        # WIF keys (bytes or string)
        self.assertEqual(ref_tuple, prvkeyinfo_from_prvkey(wif))
        self.assertEqual(ref_tuple, prvkeyinfo_from_prvkey(wif_str))
        self.assertEqual(ref_tuple, prvkeyinfo_from_prvkey(" " + wif_str + " "))

    def test_exceptions(self):

        xprv = (
            "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiC"
            "hkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
        ).encode()
        xprvd = bip32.deserialize(xprv)

        # Compressed key provided, uncompressed key requested
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, xprvd, "mainnet", False)
        # prvkeyinfo_from_prvkey(xprvd, 'mainnet', False)

        # Mainnet key provided, testnet key requested
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, xprvd, "testnet", True)
        # prvkeyinfo_from_prvkey(xprvd, 'testnet', True)

        # Compression requirement mismatch
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, xprv, "mainnet", False)
        # prvkeyinfo_from_prvkey(xprv, 'mainnet', False)

        # Mainnet key provided, testnet key requested
        self.assertRaises(ValueError, prvkeyinfo_from_prvkey, xprv, "testnet", True)
        # prvkeyinfo_from_prvkey(xprv, 'testnet', True)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import json
import unittest
from os import path

from btclib import bip39
from btclib.base58 import b58decode, b58encode
from btclib.base58address import p2pkh, p2wpkh_p2sh
from btclib.bech32address import p2wpkh
from btclib.bip32 import (
    crack_prvkey,
    derive,
    deserialize,
    mxprv_from_bip39_mnemonic,
    rootxprv_from_seed,
    serialize,
    xpub_from_xprv,
)
from btclib.curves import secp256k1 as ec
from btclib.network import NETWORKS


class TestBIP32(unittest.TestCase):
    def test_serialize(self):
        xprv = "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
        xprv_dict = deserialize(xprv)
        xprv_dict = deserialize(xprv_dict)
        xpr2 = serialize(xprv_dict)
        self.assertEqual(xpr2.decode(), xprv)

        invalid_key = (ec.n).to_bytes(ec.nsize, "big")
        # private key not in [1, n-1]
        xprv_dict["key"] = b"\x00" + invalid_key
        self.assertRaises(ValueError, deserialize, xprv_dict)
        # deserialize(xprv_dict)
        decoded_key = b58decode(xprv, 78)
        xkey = b58encode(decoded_key[:46] + invalid_key, 78)
        self.assertRaises(ValueError, deserialize, xkey)
        # deserialize(xkey)

        xpub = xpub_from_xprv(xprv)
        xpub2 = xpub_from_xprv(deserialize(xprv))
        self.assertEqual(xpub, xpub2)

    def test_utils(self):
        # root key, zero depth
        xprv = "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiChkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
        xdict = deserialize(xprv)

        decoded_key = b58decode(xprv, 78)
        self.assertEqual(xdict["version"], decoded_key[:4])
        self.assertEqual(xdict["depth"], decoded_key[4])
        self.assertEqual(xdict["parent_fingerprint"], decoded_key[5:9])
        self.assertEqual(xdict["index"], decoded_key[9:13])
        self.assertEqual(xdict["chain_code"], decoded_key[13:45])
        self.assertEqual(xdict["key"], decoded_key[45:])

        # Zero depth with non-zero parent_fingerprint b"\x01\x01\x01\x01"
        f2 = b"\x00\x00\x00\x0f"
        invalid_key = b58encode(decoded_key[:5] + f2 + decoded_key[9:], 78)
        self.assertRaises(ValueError, deserialize, invalid_key)
        # deserialize(invalid_key)

        # Zero depth with non-zero index
        i2 = b"\x00\x00\x00\xff"
        invalid_key = b58encode(decoded_key[:9] + i2 + decoded_key[13:], 78)
        self.assertRaises(ValueError, deserialize, invalid_key)
        # deserialize(invalid_key)

        # Non-zero depth (255) with zero parent_fingerprint (00000000)
        d2 = b"\xff"
        invalid_key = b58encode(decoded_key[:4] + d2 + decoded_key[5:], 78)
        self.assertRaises(ValueError, deserialize, invalid_key)
        # deserialize(invalid_key)

        child_key = derive(xprv, 0)

        # Derivation path final depth 256>255
        self.assertRaises(ValueError, derive, child_key, "." + 255 * "/0")
        # derive(child_key, "."+255*"/0")

        # Empty derivation path
        self.assertRaises(ValueError, derive, child_key, "")
        # derive(child_key, "")

        # Invalid derivation path root: ";"
        self.assertRaises(ValueError, derive, child_key, ";/0")
        # derive(child_key, ";/0")

        # Derivation path depth 256>255
        self.assertRaises(ValueError, derive, child_key, "." + 256 * "/0")
        # derive(child_key, "." + 256*"/0")

        # xkey is not a public one
        # self.assertRaises(ValueError, p2pkh, xprv)
        # p2pkh(xprv)

    def test_bip32_vectors(self):
        """BIP32 test vector 3

        https://github.com/bitcoin/bips/blob/master/bip-0032.mediawiki
        """
        fname = "bip32_test_vectors.json"
        filename = path.join(path.dirname(__file__), "test_data", fname)
        with open(filename, "r") as f:
            test_vectors = json.load(f)

        for seed in test_vectors:
            mxprv = rootxprv_from_seed(seed)
            for der_path, xpub, xprv in test_vectors[seed]:
                self.assertEqual(xprv, derive(mxprv, der_path).decode())
                self.assertEqual(xpub, xpub_from_xprv(xprv).decode())

    def test_bip39_vectors(self):
        """BIP32 test vectors from BIP39

        https://github.com/bitcoin/bips/blob/master/bip-0039.mediawiki
        """
        fname = "bip39_test_vectors.json"
        filename = path.join(path.dirname(__file__), "test_data", fname)
        with open(filename, "r") as f:
            test_vectors = json.load(f)["english"]

        for test_vector in test_vectors:
            seed = test_vector[2]
            rootxprv = rootxprv_from_seed(seed)
            self.assertEqual(rootxprv, test_vector[3].encode("ascii"))

    def test_mainnet(self):
        # bitcoin core derivation style
        rootxprv = "xprv9s21ZrQH143K2ZP8tyNiUtgoezZosUkw9hhir2JFzDhcUWKz8qFYk3cxdgSFoCMzt8E2Ubi1nXw71TLhwgCfzqFHfM5Snv4zboSebePRmLS"

        # m / 0h / 0h / 463h
        addr1 = b"1DyfBWxhVLmrJ7keyiHeMbt7N3UdeGU4G5"
        indexes = [0x80000000, 0x80000000, 0x800001CF]
        addr = p2pkh(xpub_from_xprv(derive(rootxprv, indexes)))
        self.assertEqual(addr, addr1)
        path = "m / 0h / 0h / 463h"
        addr = p2pkh(xpub_from_xprv(derive(rootxprv, path)))
        self.assertEqual(addr, addr1)

        # m / 0h / 0h / 267h
        addr2 = b"11x2mn59Qy43DjisZWQGRResjyQmgthki"
        indexes = [0x80000000, 0x80000000, 0x8000010B]
        addr = p2pkh(xpub_from_xprv(derive(rootxprv, indexes)))
        self.assertEqual(addr, addr2)
        path = "M / 0H / 0h // 267' / "
        addr = p2pkh(xpub_from_xprv(derive(rootxprv, path)))
        self.assertEqual(addr, addr2)

        seed = "bfc4cbaad0ff131aa97fa30a48d09ae7df914bcc083af1e07793cd0a7c61a03f65d622848209ad3366a419f4718a80ec9037df107d8d12c19b83202de00a40ad"
        xprv = rootxprv_from_seed(seed)
        xpub = "xpub661MyMwAqRbcFMYjmw8C6dJV97a4oLss6hb3v9wTQn2X48msQB61RCaLGtNhzgPCWPaJu7SvuB9EBSFCL43kTaFJC3owdaMka85uS154cEh"
        self.assertEqual(xpub_from_xprv(xprv).decode(), xpub)

        ind = "./0/0"
        addr = p2pkh(xpub_from_xprv(derive(xprv, ind)))
        self.assertEqual(addr, b"1FcfDbWwGs1PmyhMVpCAhoTfMnmSuptH6g")

        ind = "./0/1"
        addr = p2pkh(xpub_from_xprv(derive(xprv, ind)))
        self.assertEqual(addr, b"1K5GjYkZnPFvMDTGaQHTrVnd8wjmrtfR5x")

        ind = "./0/2"
        addr = p2pkh(xpub_from_xprv(derive(xprv, ind)))
        self.assertEqual(addr, b"1PQYX2uN7NYFd7Hq22ECMzfDcKhtrHmkfi")

        ind = "./1/0"
        addr = p2pkh(xpub_from_xprv(derive(xprv, ind)))
        self.assertEqual(addr, b"1BvSYpojWoWUeaMLnzbkK55v42DbizCoyq")

        ind = "./1/1"
        addr = p2pkh(xpub_from_xprv(derive(xprv, ind)))
        self.assertEqual(addr, b"1NXB59hF4QzYpFrB7o6usLBjbk2D3ZqxAL")

        ind = "./1/2"
        addr = p2pkh(xpub_from_xprv(derive(xprv, ind)))
        self.assertEqual(addr, b"16NLYkKtvYhW1Jp86tbocku3gxWcvitY1w")

        # version/key mismatch in extended parent key
        temp = b58decode(rootxprv)
        bad_xprv = b58encode(temp[0:45] + b"\x01" + temp[46:], 78)
        self.assertRaises(ValueError, derive, bad_xprv, 1)
        # derive(bad_xprv, 1)

        # version/key mismatch in extended parent key
        xpub = xpub_from_xprv(rootxprv)
        temp = b58decode(xpub)
        bad_xpub = b58encode(temp[0:45] + b"\x00" + temp[46:], 78)
        self.assertRaises(ValueError, derive, bad_xpub, 1)
        # derive(bad_xpub, 1)

        # no private/hardened derivation from pubkey
        self.assertRaises(ValueError, derive, xpub, 0x80000000)
        # derive(xpub, 0x80000000)

    def test_testnet(self):
        # bitcoin core derivation style
        rootxprv = "tprv8ZgxMBicQKsPe3g3HwF9xxTLiyc5tNyEtjhBBAk29YA3MTQUqULrmg7aj9qTKNfieuu2HryQ6tGVHse9x7ANFGs3f4HgypMc5nSSoxwf7TK"

        # m / 0h / 0h / 51h
        addr1 = b"mfXYCCsvWPgeCv8ZYGqcubpNLYy5nYHbbj"
        indexes = [0x80000000, 0x80000000, 0x80000000 + 51]
        addr = p2pkh(xpub_from_xprv(derive(rootxprv, indexes)))
        self.assertEqual(addr, addr1)
        path = "m/0h/0h/51h"
        addr = p2pkh(xpub_from_xprv(derive(rootxprv, path)))
        self.assertEqual(addr, addr1)

        # m / 0h / 1h / 150h
        addr2 = b"mfaUnRFxVvf55uD1P3zWXpprN1EJcKcGrb"
        indexes = [0x80000000, 0x80000000 + 1, 0x80000000 + 150]
        addr = p2pkh(xpub_from_xprv(derive(rootxprv, indexes)))
        self.assertEqual(addr, addr2)
        path = "m/0h/1h/150h"
        addr = p2pkh(xpub_from_xprv(derive(rootxprv, path)))
        self.assertEqual(addr, addr2)

    def test_exceptions(self):
        # valid xprv
        xprv = "xprv9s21ZrQH143K2oxHiQ5f7D7WYgXD9h6HAXDBuMoozDGGiYHWsq7TLBj2yvGuHTLSPCaFmUyN1v3fJRiY2A4YuNSrqQMPVLZKt76goL6LP7L"

        # invalid index
        self.assertRaises(ValueError, derive, xprv, "invalid index")
        # derive(xprv, "invalid index")

        # a 4 bytes int is required, not 3
        self.assertRaises(ValueError, derive, xprv, "800000")
        # derive(xprv, "800000")

        # Invalid derivation path root: ""
        self.assertRaises(ValueError, derive, xprv, "/1")
        # derive(xprv, "/1")

        # invalid checksum
        xprv = "xppp9s21ZrQH143K2oxHiQ5f7D7WYgXD9h6HAXDBuMoozDGGiYHWsq7TLBj2yvGuHTLSPCaFmUyN1v3fJRiY2A4YuNSrqQMPVLZKt76goL6LP7L"
        self.assertRaises(ValueError, derive, xprv, 0x80000000)
        # derive(xprv, 0x80000000)

        # invalid extended key version
        version = b"\x04\x88\xAD\xE5"
        xkey = version + b"\x00" * 74
        xkey = b58encode(xkey, 78)
        self.assertRaises(ValueError, derive, xkey, 0x80000000)
        # derive(xkey, 0x80000000)

        # unknown extended key version
        version = b"\x04\x88\xAD\xE5"
        seed = "5b56c417303faa3fcba7e57400e120a0ca83ec5a4fc9ffba757fbe63fbd77a89a1a3be4c67196f57c39a88b76373733891bfaba16ed27a813ceed498804c0570"
        self.assertRaises(ValueError, rootxprv_from_seed, seed, version)
        # rootxprv_from_seed(seed, version)

        # extended key is not a private one
        xpub = "xpub6H1LXWLaKsWFhvm6RVpEL9P4KfRZSW7abD2ttkWP3SSQvnyA8FSVqNTEcYFgJS2UaFcxupHiYkro49S8yGasTvXEYBVPamhGW6cFJodrTHy"
        self.assertRaises(ValueError, xpub_from_xprv, xpub)
        # xpub_from_xprv(xpub)

        # Absolute derivation path for non-master key
        self.assertRaises(ValueError, derive, xpub, "m/44h/0h/1h/0/10")
        # derive(xpub, "m/0/1")

        # empty derivation path
        self.assertRaises(ValueError, derive, xpub, "")
        # derive(xpub, "")

        # extended key is not a public one
        self.assertRaises(ValueError, p2pkh, xprv)
        # p2pkh(xprv)

    def test_exceptions2(self):
        rootxprv = "xprv9s21ZrQH143K2ZP8tyNiUtgoezZosUkw9hhir2JFzDhcUWKz8qFYk3cxdgSFoCMzt8E2Ubi1nXw71TLhwgCfzqFHfM5Snv4zboSebePRmLS"
        d = deserialize(rootxprv)
        self.assertEqual(serialize(d).decode(), rootxprv)

        # invalid 34-bytes key length
        d["key"] += b"\x00"
        self.assertRaises(ValueError, serialize, d)
        # serialize(d)

        # invalid 33-bytes chain_code length
        d = deserialize(rootxprv)
        d["chain_code"] += b"\x00"
        self.assertRaises(ValueError, serialize, d)
        # serialize(d)

        # invalid 5-bytes parent_fingerprint length
        d = deserialize(rootxprv)
        d["parent_fingerprint"] += b"\x00"
        self.assertRaises(ValueError, serialize, d)
        # serialize(d)

        # invalid 5-bytes index length
        d = deserialize(rootxprv)
        d["index"] += b"\x00"
        self.assertRaises(ValueError, serialize, d)
        # serialize(d)

        # invalid depth (256)
        d = deserialize(rootxprv)
        d["depth"] = 256
        self.assertRaises(ValueError, serialize, d)
        # serialize(d)

        # zero depth with non-zero index b"\x00\x00\x00\x01"
        d = deserialize(rootxprv)
        d["index"] = b"\x00\x00\x00\x01"
        self.assertRaises(ValueError, serialize, d)
        # serialize(d)

        # zero depth with non-zero parent_fingerprint b"\x00\x00\x00\x01"
        d = deserialize(rootxprv)
        d["parent_fingerprint"] = b"\x00\x00\x00\x01"
        self.assertRaises(ValueError, serialize, d)
        # serialize(d)

        # non-zero depth (1) with zero parent_fingerprint b"\x00\x00\x00\x00"
        xprv = deserialize(derive(rootxprv, 1))
        xprv["parent_fingerprint"] = b"\x00\x00\x00\x00"
        self.assertRaises(ValueError, serialize, xprv)
        # serialize(xprv)

        # int too big to convert
        self.assertRaises(OverflowError, derive, rootxprv, 256 ** 4)

        # Index must be 4-bytes, not 5
        self.assertRaises(ValueError, derive, rootxprv, b"\x00" * 5)
        # derive(rootxprv, b"\x00"*5)

    def test_testnet_versions(self):

        # data cross-checked with Electrum and
        # https://jlopp.github.io/xpub-converter/

        # 128 bits
        raw_entr = bytes.fromhex("6" * 32)
        # 12 words
        mnemonic = bip39.mnemonic_from_entropy(raw_entr, "en")
        seed = bip39.seed_from_mnemonic(mnemonic, "")

        # p2pkh BIP44
        # m / 44h / coin_typeh / accounth / change / address_index
        path = "m/44h/1h/0h"
        version = NETWORKS["testnet"]["bip32_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "tpubDChqWo2Xi2wNsxyJBE8ipcTJHLKWcqeeNUKBVTpUCNPZkHzHTm3qKAeHqgCou1t8PAY5ZnJ9QDa6zXSZxmjDnhiBpgZ7f6Yv88wEm5HXVbm"
        self.assertEqual(xpub.decode(), exp)
        # first addresses
        xpub_ext = derive(xpub, "./0/0")  # external
        address = p2pkh(xpub_ext)
        exp_address = b"moutHSzeFWViMNEcvBxKzNCMj2kca8MvE1"
        self.assertEqual(address, exp_address)
        xpub_int = derive(xpub, "./1/0")  # internal
        address = p2pkh(xpub_int)
        exp_address = b"myWcXdNais9ExumnGKnNoJwoihQKfNPG9i"
        self.assertEqual(address, exp_address)

        # legacy segwit (p2wsh-p2sh)
        # m / 49h / coin_typeh / accounth / change / address_index
        path = "m/49h/1h/0h"
        version = NETWORKS["testnet"]["slip32_p2wsh_p2sh_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "upub5Dj8j7YrwodV68mt58QmNpSzjqjso2WMXEpLGLSvskKccGuXhCh3dTedkzVLAePA617UyXAg2vdswJXTYjU4qjMJaHU79GJVVJCAiy9ezZ2"
        self.assertEqual(xpub.decode(), exp)
        # first addresses
        xpub_ext = derive(xpub, "./0/0")  # external
        address = p2wpkh_p2sh(xpub_ext)
        exp_address = b"2Mw8tQ6uT6mHhybarVhjgomUhHQJTeV9A2c"
        self.assertEqual(address, exp_address)
        xpub_int = derive(xpub, "./1/0")  # internal
        address = p2wpkh_p2sh(xpub_int)
        exp_address = b"2N872CRJ3E1CzWjfixXr3aeC3hkF5Cz4kWb"
        self.assertEqual(address, exp_address)

        # legacy segwit (p2wsh-p2sh)
        # m / 49h / coin_typeh / accounth / change / address_index
        path = "m/49h/1h/0h"
        version = NETWORKS["testnet"]["slip32_p2wpkh_p2sh_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "Upub5QdDrMHJWmBrWhwG1nskCtnoTdn91PBwqWU1BbiUFXA2ETUSTc5KiaWZZhSoj5c4KUBTr7Anv92P4U9Dqxd1zDTyQkaWYfmVP2U3Js1W5cG"
        self.assertEqual(xpub.decode(), exp)

        # native segwit (p2wpkh)
        # m / 84h / coin_typeh / accounth / change / address_index
        path = "m/84h/1h/0h"
        version = NETWORKS["testnet"]["slip32_p2wpkh_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "vpub5ZhJmduYY7M5J2qCJgSW7hunX6zJrr5WuNg2kKt321HseZEYxqJc6Zso47aNXQw3Wf3sA8kppbfsxnLheUNXcL3xhzeBHLNp8fTVBN6DnJF"
        self.assertEqual(xpub.decode(), exp)
        # first addresses
        xpub_ext = derive(xpub, "./0/0")  # external
        # explicit network is required to discriminate from testnet
        address = p2wpkh(xpub_ext, "regtest")
        exp_address = b"bcrt1qv8lcnmj09rpdqwgl025h2deygur64z4hqf7me5"
        self.assertEqual(address, exp_address)
        xpub_int = derive(xpub, "./1/0")  # internal
        # explicit network is required to discriminate from testnet
        address = p2wpkh(xpub_int, "regtest")
        exp_address = b"bcrt1qqhxvky4y6qkwpvdzqjkdafmj20vs5trmt6y8w5"
        self.assertEqual(address, exp_address)

        # native segwit (p2wsh)
        # m / 84h / coin_typeh / accounth / change / address_index
        path = "m/84h/1h/0h"
        version = NETWORKS["testnet"]["slip32_p2wsh_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "Vpub5kbPtsdz74uSibzaFLuUwnFbEu2a5Cm7DeKhfb9aPn8HGjoTjEgtBgjirpXr5r9wk87r2ikwhp4P5wxTwhXUkpAdYTkagjqp2PjMmGPBESU"
        self.assertEqual(xpub.decode(), exp)

    def test_mainnet_versions(self):

        # data cross-checked with Electrum and
        # https://jlopp.github.io/xpub-converter/

        # 128 bits
        raw_entr = bytes.fromhex("6" * 32)
        # 12 words
        mnemonic = bip39.mnemonic_from_entropy(raw_entr, "en")
        seed = bip39.seed_from_mnemonic(mnemonic, "")

        # p2pkh BIP44
        # m / 44h / coin_typeh / accounth / change / address_index
        path = "m/44h/0h/0h"
        version = NETWORKS["mainnet"]["bip32_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "xpub6C3uWu5Go5q62JzJpbjyCLYRGLYvexFeiepZTsYZ6SRexARkNfjG7GKtQVuGR3KHsyKsAwv7Hz3iNucPp6pfHiLvBczyK1j5CtBtpHB3NKx"
        self.assertEqual(xpub.decode(), exp)
        # first addresses
        xpub_ext = derive(xpub, "./0/0")  # external
        address = p2pkh(xpub_ext)
        exp_address = b"1DDKKVHoFWGfctyEEJvrusqq6ipEaieGCq"
        self.assertEqual(address, exp_address)
        xpub_int = derive(xpub, "./1/0")  # internal
        address = p2pkh(xpub_int)
        exp_address = b"1FhKoffreKHzhtBMVW9NSsg3ZF148JPGoR"
        self.assertEqual(address, exp_address)

        # legacy segwit (p2wsh-p2sh)
        # m / 49h / coin_typeh / accounth / change / address_index
        path = "m/49h/0h/0h"
        version = NETWORKS["mainnet"]["slip32_p2wsh_p2sh_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "ypub6YBGdYufCVeoPVmNXfdrWhaBCXsQoLKNetNmD9bPTrKmnKVmiyU8f1uJqwGdmBb8kbAZpHoYfXQTLbWpkXc4skQDAreeCUXdbX9k8vtiHsN"
        self.assertEqual(xpub.decode(), exp)
        # first addresses
        xpub_ext = derive(xpub, "./0/0")  # external
        address = p2wpkh_p2sh(xpub_ext)
        exp_address = b"3FmNAiTCWe5kPMgc4dtSgEdY8VuaCiJEH8"
        self.assertEqual(address, exp_address)
        xpub_int = derive(xpub, "./1/0")  # internal
        address = p2wpkh_p2sh(xpub_int)
        exp_address = b"34FLgkoRYX5Q5fqiZCZDwsK5GpXxmFuLJN"
        self.assertEqual(address, exp_address)

        # legacy segwit (p2wpkh-p2sh)
        # m / 49h / coin_typeh / accounth / change / address_index
        path = "m/49h/0h/0h"
        version = NETWORKS["mainnet"]["slip32_p2wpkh_p2sh_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "Ypub6j5Mkne6mTDAp4vkUL6qLmuyvKug1gzxyA2S8QrvqdABQW4gVNrQk8mEeeE7Kcp2z4EYgsofYjnxTm8b3km22EWt1Km3bszdVFRcipc6rXu"
        self.assertEqual(xpub.decode(), exp)

        # native segwit (p2wpkh)
        # m / 84h / coin_typeh / accounth / change / address_index
        path = "m/84h/0h/0h"
        version = NETWORKS["mainnet"]["slip32_p2wpkh_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "zpub6qg3Uc1BAQkQvcBUYMmZHSzbsshSon3FvJ8yvH3ZZMjFNvJkwSji8UUwghiF3wvpvSvcNWVP8kfUhc2V2RwGp6pTC3ouj6njj956f26TniN"
        self.assertEqual(xpub.decode(), exp)
        # first addresses
        xpub_ext = derive(xpub, "./0/0")  # external
        address = p2wpkh(xpub_ext)
        exp_address = b"bc1q0hy024867ednvuhy9en4dggflt5w9unw4ztl5a"
        self.assertEqual(address, exp_address)
        xpub_int = derive(xpub, "./1/0")  # internal
        address = p2wpkh(xpub_int)
        exp_address = b"bc1qy4x03jyl88h2zeg7l287xhv2xrwk4c3ztfpjd2"
        self.assertEqual(address, exp_address)

        # native segwit (p2wsh)
        # m / 84h / coin_typeh / accounth / change / address_index
        path = "m/84h/0h/0h"
        version = NETWORKS["mainnet"]["slip32_p2wsh_prv"]
        rootprv = rootxprv_from_seed(seed, version)
        xprv = derive(rootprv, path)
        xpub = xpub_from_xprv(xprv)
        exp = "Zpub72a8bqjcjNJnMBLrV2EY7XLQbfji28irEZneqYK6w8Zf16sfhr7zDbLsVQficP9j9uzbF6VW1y3ypmeFKf6Dxaw82WvK8WFjcsLyEvMNZjF"
        self.assertEqual(xpub.decode(), exp)

    def test_rootxprv_from_mnemonic(self):
        mnemonic = (
            "abandon abandon atom  trust  ankle   walnut  "
            "oil     across  awake bunker divorce abstract"
        )
        passphrase = ""
        rootxprv = mxprv_from_bip39_mnemonic(mnemonic, passphrase)
        exp = "xprv9s21ZrQH143K3ZxBCax3Wu25iWt3yQJjdekBuGrVa5LDAvbLeCT99U59szPSFdnMe5szsWHbFyo8g5nAFowWJnwe8r6DiecBXTVGHG124G1"
        self.assertEqual(rootxprv.decode(), exp)

    def test_crack(self):
        parent_xpub = "xpub6BabMgRo8rKHfpAb8waRM5vj2AneD4kDMsJhm7jpBDHSJvrFAjHJHU5hM43YgsuJVUVHWacAcTsgnyRptfMdMP8b28LYfqGocGdKCFjhQMV"
        child_xprv = "xprv9xkG88dGyiurKbVbPH1kjdYrA8poBBBXa53RKuRGJXyruuoJUDd8e4m6poiz7rV8Z4NoM5AJNcPHN6aj8wRFt5CWvF8VPfQCrDUcLU5tcTm"
        parent_xprv = crack_prvkey(parent_xpub, child_xprv)
        self.assertEqual(xpub_from_xprv(parent_xprv).decode(), parent_xpub)
        # same check with XKeyDict
        parent_xprv = crack_prvkey(deserialize(parent_xpub), deserialize(child_xprv))
        self.assertEqual(xpub_from_xprv(parent_xprv).decode(), parent_xpub)

        # extended parent key is not a public one
        self.assertRaises(ValueError, crack_prvkey, parent_xprv, child_xprv)
        # crack_prvkey(parent_xprv, child_xprv)

        # extended child key is not a private one
        self.assertRaises(ValueError, crack_prvkey, parent_xpub, parent_xpub)
        # crack_prvkey(parent_xpub, parent_xpub)

        # wrong child/parent depth relation
        child_xpub = xpub_from_xprv(child_xprv)
        self.assertRaises(ValueError, crack_prvkey, child_xpub, child_xprv)
        # crack_prvkey(child_xpub, child_xprv)

        # not a child for the provided parent
        child0_xprv = derive(parent_xprv, 0)
        grandchild_xprv = derive(child0_xprv, 0)
        self.assertRaises(ValueError, crack_prvkey, child_xpub, grandchild_xprv)
        # crack_prvkey(child_xpub, grandchild_xprv)

        # hardened derivation
        hardened_child_xprv = derive(parent_xprv, 0x80000000)
        self.assertRaises(ValueError, crack_prvkey, parent_xpub, hardened_child_xprv)
        # crack_prvkey(parent_xpub, hardened_child_xprv)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest
from hashlib import sha1 as hf

from btclib import dh
from btclib.curvemult import mult
from btclib.curves import CURVES
from btclib.secpoint import bytes_from_point

ec = CURVES["secp160r1"]


class TestEcdh(unittest.TestCase):
    def test_ecdh(self):
        size = 20

        dU = 0x1
        QU = mult(dU, ec.G, ec)
        dV = 0x2
        QV = mult(dV, ec.G, ec)

        keyingdataU = dh.diffie_hellman(dh.ansi_x963_kdf, dU, QV, size, ec, hf)
        keyingdataV = dh.diffie_hellman(dh.ansi_x963_kdf, dV, QU, size, ec, hf)
        self.assertEqual(keyingdataU, keyingdataV)

    def test_key_deployment(self):
        """GEC 2: Test Vectors for SEC 1, section 4.1

        http://read.pudn.com/downloads168/doc/772358/TestVectorsforSEC%201-gec2.pdf
        """

        # 4.1.1
        # ec = secp160r1
        # hf = sha1

        # 4.1.2
        dU = 971761939728640320549601132085879836204587084162
        self.assertEqual(
            format(dU, str(ec.nsize) + "x"), "aa374ffc3ce144e6b073307972cb6d57b2a4e982"
        )
        QU = mult(dU, ec.G, ec)
        self.assertEqual(
            QU,
            (
                466448783855397898016055842232266600516272889280,
                1110706324081757720403272427311003102474457754220,
            ),
        )
        self.assertEqual(
            bytes_from_point(QU, ec).hex(), "0251b4496fecc406ed0e75a24a3c03206251419dc0"
        )

        # 4.1.3
        dV = 399525573676508631577122671218044116107572676710
        self.assertEqual(
            format(dV, str(ec.nsize) + "x"), "45fb58a92a17ad4b15101c66e74f277e2b460866"
        )
        QV = mult(dV, ec.G, ec)
        self.assertEqual(
            QV,
            (
                420773078745784176406965940076771545932416607676,
                221937774842090227911893783570676792435918278531,
            ),
        )
        self.assertEqual(
            bytes_from_point(QV, ec).hex(), "0349b41e0e9c0369c2328739d90f63d56707c6e5bc"
        )

        # expected results
        z_exp = 1155982782519895915997745984453282631351432623114
        zstr = "ca7c0f8c3ffa87a96e1b74ac8e6af594347bb40a"
        size = 20
        keying_data_exp = "744ab703f5bc082e59185f6d049d2d367db245c2"

        # 4.1.4
        z, _ = mult(dU, QV, ec)  # x coordinate only
        self.assertEqual(z, z_exp)
        self.assertEqual(format(z, str(ec.psize) + "x"), zstr)
        keyingdata = dh.ansi_x963_kdf(z.to_bytes(ec.psize, "big"), size, ec, hf)
        self.assertEqual(keyingdata.hex(), keying_data_exp)

        # 4.1.5
        z, _ = mult(dV, QU, ec)  # x coordinate only
        self.assertEqual(z, z_exp)
        self.assertEqual(format(z, str(ec.psize) + "x"), zstr)
        keyingdata = dh.ansi_x963_kdf(z.to_bytes(ec.psize, "big"), size, ec, hf)
        self.assertEqual(keyingdata.hex(), keying_data_exp)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest
from hashlib import sha256, sha384

from btclib import pedersen
from btclib.curves import CURVES, secp256k1

secp256r1 = CURVES["secp256r1"]
secp384r1 = CURVES["secp384r1"]


def test_second_generator():
    """
    important remarks on secp256-zkp prefix for
    compressed encoding of the second generator:
    https://github.com/garyyu/rust-secp256k1-zkp/wiki/Pedersen-Commitment
    """

    H = (
        0x50929B74C1A04954B78B4B6035E97A5E078A5A0F28EC96D547BFEE9ACE803AC0,
        0x31D3C6863973926E049E637CB1B5F40A36DAC28AF1766968C30C2313F3A38904,
    )
    assert H == pedersen.second_generator(secp256k1, sha256)

    H = pedersen.second_generator(secp256r1, sha256)
    H = pedersen.second_generator(secp384r1, sha384)


class TestPedersenCommitment(unittest.TestCase):
    def test_commitment(self):

        ec = secp256k1
        hf = sha256

        r1 = 0x1
        v1 = 0x2
        # r1*G + v1*H
        C1 = pedersen.commit(r1, v1, ec, hf)
        self.assertTrue(pedersen.open(r1, v1, C1, ec, hf))

        r2 = 0x3
        v2 = 0x4
        # r2*G + v2*H
        C2 = pedersen.commit(r2, v2, ec, hf)
        self.assertTrue(pedersen.open(r2, v2, C2, ec, hf))

        # Pedersen Commitment is additively homomorphic
        # Commit(r1, v1) + Commit(r2, v2) = Commit(r1+r2, v1+r2)
        R = pedersen.commit(r1 + r2, v1 + v2, ec, hf)
        self.assertEqual(ec.add(C1, C2), R)

        # commit does not open (with catched exception)
        self.assertFalse(pedersen.open((r1, r1), v1, C2, ec, hf))



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import copy
import unittest

from btclib import bip32
from btclib.alias import INF
from btclib.base58 import b58encode
from btclib.base58wif import wif_from_prvkey
from btclib.curvemult import mult
from btclib.curves import CURVES
from btclib.network import NETWORKS
from btclib.secpoint import bytes_from_point
from btclib.to_prvkey import _prvkeyinfo_from_xprv
from btclib.to_pubkey import (
    _pubkeyinfo_from_xpub,
    fingerprint,
    point_from_key,
    point_from_pubkey,
    pubkeyinfo_from_key,
    pubkeyinfo_from_pubkey,
)

secp256r1 = CURVES["secp256r1"]


class TestToPubKey(unittest.TestCase):
    def test_point_from_key(self):

        # prvkeys
        xprv = (
            "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiC"
            "hkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
        ).encode()
        xprv_str = xprv.decode("ascii")
        xprv_dict = bip32.deserialize(xprv)
        q, network, _ = _prvkeyinfo_from_xprv(xprv)
        ec = NETWORKS[network]["curve"]
        q_bytes = q.to_bytes(ec.nsize, "big")
        wif1 = wif_from_prvkey(q, network, True)
        wif2 = wif_from_prvkey(q, network, False)

        # pubkeys
        Q = mult(q, ec.G, ec)
        xpub = bip32.xpub_from_xprv(xprv)
        xpub_str = xpub.decode("ascii")
        xpub_dict = bip32.deserialize(xpub)
        Q_compr = xpub_dict["key"]
        Q_compr_hexstr = Q_compr.hex()
        Q_uncompr = bytes_from_point(Q, ec, False)
        Q_uncompr_hexstr = Q_uncompr.hex()

        # int prvkey
        self.assertEqual(point_from_key(q, ec), Q)
        self.assertEqual(point_from_key(q_bytes, ec), Q)

        # wif prvkey
        self.assertEqual(point_from_key(wif1, ec), Q)
        self.assertEqual(point_from_key(wif2, ec), Q)

        # BIP32 prvkey
        self.assertEqual(point_from_key(xprv, ec), Q)
        self.assertEqual(point_from_key(xprv_str, ec), Q)
        self.assertEqual(point_from_key(" " + xprv_str + " ", ec), Q)
        self.assertEqual(point_from_key(xprv_dict, ec), Q)

        # BIP32 pubkey
        self.assertEqual(point_from_key(xpub, ec), Q)
        self.assertEqual(point_from_key(xpub_str, ec), Q)
        self.assertEqual(point_from_key(" " + xpub_str + " ", ec), Q)
        self.assertEqual(point_from_key(xpub_dict, ec), Q)

        # compressed SEC Octets (bytes or hex-string, with 02 or 03 prefix)
        self.assertEqual(point_from_key(Q_compr, ec), Q)
        self.assertRaises(ValueError, point_from_key, b"\x00" + Q_compr, ec)
        self.assertEqual(point_from_key(Q_compr_hexstr, ec), Q)
        self.assertEqual(point_from_key(" " + Q_compr_hexstr + " ", ec), Q)
        self.assertRaises(ValueError, point_from_key, Q_compr_hexstr + "00", ec)

        # uncompressed SEC Octets (bytes or hex-string, with 04 prefix)
        self.assertEqual(point_from_key(Q_uncompr, ec), Q)
        self.assertRaises(ValueError, point_from_key, b"\x00" + Q_uncompr, ec)
        self.assertEqual(point_from_key(Q_uncompr_hexstr, ec), Q)
        self.assertEqual(point_from_key(" " + Q_uncompr_hexstr + " ", ec), Q)
        self.assertRaises(ValueError, point_from_key, Q_uncompr_hexstr + "00", ec)

        # native tuple
        self.assertEqual(point_from_key(Q, ec), Q)

        # Invalid point: 7 is not a field element
        Q = INF
        self.assertRaises(ValueError, point_from_key, Q, ec)
        Q_compr = b"\x02" + Q[0].to_bytes(ec.psize, "big")
        self.assertRaises(ValueError, point_from_key, Q_compr, ec)
        Q_uncompr = (
            b"\x04" + Q[0].to_bytes(ec.psize, "big") + Q[1].to_bytes(ec.psize, "big")
        )
        self.assertRaises(ValueError, point_from_key, Q_uncompr, ec)
        Q_compr_hexstr = Q_compr.hex()
        self.assertRaises(ValueError, point_from_key, Q_compr_hexstr, ec)
        Q_uncompr_hexstr = Q_uncompr.hex()
        self.assertRaises(ValueError, point_from_key, Q_uncompr_hexstr, ec)
        t = xpub_dict["version"]
        t += xpub_dict["depth"].to_bytes(1, "big")
        t += xpub_dict["parent_fingerprint"]
        t += xpub_dict["index"]
        t += xpub_dict["chain_code"]
        t += Q_compr
        xpub = b58encode(t, 78)
        self.assertRaises(ValueError, point_from_key, xpub, ec)
        xpub_str = xpub.decode("ascii")
        self.assertRaises(ValueError, point_from_key, xpub_str, ec)

        # pubkey input
        self.assertRaises(ValueError, point_from_pubkey, xprv, ec)
        self.assertRaises(ValueError, point_from_pubkey, xprv_dict, ec)

    def test_pubkeyinfo_from_key(self):

        # prvkeys
        xprv = (
            "xprv9s21ZrQH143K3QTDL4LXw2F7HEK3wJUD2nW2nRk4stbPy6cq3jPPqjiC"
            "hkVvvNKmPGJxWUtg6LnF5kejMRNNU3TGtRBeJgk33yuGBxrMPHi"
        ).encode()
        xprv_str = xprv.decode("ascii")
        xprv_dict = bip32.deserialize(xprv)
        q, network, _ = _prvkeyinfo_from_xprv(xprv)
        ec = NETWORKS[network]["curve"]
        q_bytes = q.to_bytes(ec.nsize, "big")
        wif1 = wif_from_prvkey(q, network, True)
        wif2 = wif_from_prvkey(q, network, False)

        # pubkeys
        Q = mult(q, ec.G, ec)
        xpub = bip32.xpub_from_xprv(xprv)
        xpub_str = xpub.decode("ascii")
        xpub_dict = bip32.deserialize(xpub)
        Q_compr = xpub_dict["key"]
        Q_compr_hexstr = Q_compr.hex()
        Q_uncompr = bytes_from_point(Q, ec, False)
        Q_uncompr_hexstr = Q_uncompr.hex()

        # int prvkey, compressed result
        self.assertEqual(pubkeyinfo_from_key(q)[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(q_bytes)[0], Q_compr)

        # int prvkey, uncompressed result
        self.assertEqual(pubkeyinfo_from_key(q, compressed=False)[0], Q_uncompr)
        self.assertEqual(pubkeyinfo_from_key(q_bytes, compressed=False)[0], Q_uncompr)

        # compressed wif prvkey, both results
        self.assertEqual(pubkeyinfo_from_key(wif1)[0], Q_compr)
        self.assertRaises(ValueError, pubkeyinfo_from_key, wif1, compressed=False)

        # uncompressed wif prvkey, both results
        self.assertRaises(ValueError, pubkeyinfo_from_key, wif2, compressed=True)
        self.assertEqual(pubkeyinfo_from_key(wif2)[0], Q_uncompr)

        # (compressed) BIP32 prvkey, compressed results
        self.assertEqual(pubkeyinfo_from_key(xprv)[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(xprv_str)[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(" " + xprv_str + " ")[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(xprv_dict)[0], Q_compr)

        # (compressed) BIP32 prvkey, uncompressed result
        self.assertRaises(ValueError, pubkeyinfo_from_key, xprv, compressed=False)
        self.assertRaises(ValueError, pubkeyinfo_from_key, xprv_str, compressed=False)
        self.assertRaises(
            ValueError, pubkeyinfo_from_key, " " + xprv_str + " ", compressed=False
        )
        self.assertRaises(ValueError, pubkeyinfo_from_key, xprv_dict, compressed=False)

        # (compressed) BIP32 pubkey, compressed results
        self.assertEqual(pubkeyinfo_from_key(xpub)[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(xpub_str)[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(" " + xpub_str + " ")[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(xpub_dict)[0], Q_compr)

        # (compressed) BIP32 pubkey, uncompressed result
        self.assertRaises(ValueError, pubkeyinfo_from_key, xpub, compressed=False)
        self.assertRaises(ValueError, pubkeyinfo_from_key, xpub_str, compressed=False)
        self.assertRaises(
            ValueError, pubkeyinfo_from_key, " " + xpub_str + " ", compressed=False
        )
        self.assertRaises(ValueError, pubkeyinfo_from_key, xpub_dict, compressed=False)

        # compressed SEC Octets (pubkey), compressed results
        self.assertEqual(pubkeyinfo_from_key(Q_compr)[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(Q_compr_hexstr)[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(" " + Q_compr_hexstr + " ")[0], Q_compr)

        # compressed SEC Octets (pubkey), uncompressed results
        self.assertRaises(ValueError, pubkeyinfo_from_key, Q_compr, compressed=False)
        self.assertRaises(
            ValueError, pubkeyinfo_from_key, Q_compr_hexstr, compressed=False
        )
        self.assertRaises(
            ValueError,
            pubkeyinfo_from_key,
            " " + Q_compr_hexstr + " ",
            compressed=False,
        )
        self.assertRaises(ValueError, pubkeyinfo_from_key, b"\x00" + Q_compr)
        self.assertRaises(ValueError, pubkeyinfo_from_key, Q_compr_hexstr + "00")

        # uncompressed SEC Octets (pubkey), uncompressed results
        self.assertEqual(pubkeyinfo_from_key(Q_uncompr)[0], Q_uncompr)
        self.assertEqual(pubkeyinfo_from_key(Q_uncompr_hexstr)[0], Q_uncompr)
        self.assertEqual(
            pubkeyinfo_from_key(" " + Q_uncompr_hexstr + " ")[0], Q_uncompr
        )

        # uncompressed SEC Octets (pubkey), compressed results
        self.assertRaises(ValueError, pubkeyinfo_from_key, Q_uncompr, compressed=True)
        self.assertRaises(
            ValueError, pubkeyinfo_from_key, Q_uncompr_hexstr, compressed=True
        )
        self.assertRaises(
            ValueError,
            pubkeyinfo_from_key,
            " " + Q_uncompr_hexstr + " ",
            compressed=True,
        )
        self.assertRaises(ValueError, pubkeyinfo_from_key, b"\x00" + Q_uncompr)
        self.assertRaises(ValueError, pubkeyinfo_from_key, Q_uncompr_hexstr + "00")

        # native tuple input, both results
        self.assertEqual(pubkeyinfo_from_key(Q)[0], Q_compr)
        self.assertEqual(pubkeyinfo_from_key(Q, compressed=False)[0], Q_uncompr)

        # pubkeyinfo_from_pubkey does not accept prvkey inputs
        self.assertRaises(TypeError, pubkeyinfo_from_pubkey, q)
        self.assertRaises(ValueError, pubkeyinfo_from_pubkey, q_bytes)
        self.assertRaises(ValueError, pubkeyinfo_from_pubkey, xprv)
        self.assertRaises(ValueError, pubkeyinfo_from_pubkey, xprv_str)
        self.assertRaises(ValueError, pubkeyinfo_from_pubkey, xprv_dict)
        self.assertRaises(ValueError, pubkeyinfo_from_pubkey, wif1)
        self.assertRaises(ValueError, pubkeyinfo_from_pubkey, wif2)

        # Not a public key:
        xpub_dict_bad = copy.copy(xpub_dict)
        xpub_dict_bad["key"] = b"\x00" + xpub_dict["key"][1:]
        self.assertRaises(ValueError, _pubkeyinfo_from_xpub, xpub_dict_bad)
        # _pubkeyinfo_from_xpub(xpub_dict_bad)

        # Invalid point: 7 is not a field element
        Q = INF
        self.assertRaises(ValueError, pubkeyinfo_from_key, Q)
        Q_compr = b"\x02" + Q[0].to_bytes(ec.psize, "big")
        self.assertRaises(ValueError, pubkeyinfo_from_key, Q_compr)
        Q_uncompr = (
            b"\x04" + Q[0].to_bytes(ec.psize, "big") + Q[1].to_bytes(ec.psize, "big")
        )
        self.assertRaises(ValueError, pubkeyinfo_from_key, Q_uncompr)
        Q_compr_hexstr = Q_compr.hex()
        self.assertRaises(ValueError, pubkeyinfo_from_key, Q_compr_hexstr)
        Q_uncompr_hexstr = Q_uncompr.hex()
        self.assertRaises(ValueError, pubkeyinfo_from_key, Q_uncompr_hexstr)
        t = xpub_dict["version"]
        t += xpub_dict["depth"].to_bytes(1, "big")
        t += xpub_dict["parent_fingerprint"]
        t += xpub_dict["index"]
        t += xpub_dict["chain_code"]
        t += Q_compr
        xpub = b58encode(t, 78)
        self.assertRaises(ValueError, pubkeyinfo_from_key, xpub)
        xpub_str = xpub.decode("ascii")
        self.assertRaises(ValueError, pubkeyinfo_from_key, xpub_str)

    def test_fingerprint(self):
        xpub = (
            "xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2g"
            "Z29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8"
        ).encode()
        pf = fingerprint(xpub)
        # bytes are used to increase code coverage
        # dict is used to increase code coverage
        xpubd = bip32.deserialize(xpub)
        child_key = bip32.derive(xpubd, b"\x00" * 4)
        pf2 = bip32.deserialize(child_key)["parent_fingerprint"]
        self.assertEqual(pf, pf2)

    def test_exceptions(self):

        # Not a key for (testnet) network
        xpub = (
            "xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2g"
            "Z29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8"
        ).encode()
        xpubd = bip32.deserialize(xpub)
        self.assertRaises(ValueError, _pubkeyinfo_from_xpub, xpubd, "testnet", None)
        # _pubkeyinfo_from_xpub(xpubd, 'testnet', None)

        self.assertRaises(ValueError, _pubkeyinfo_from_xpub, xpubd, None, False)
        # _pubkeyinfo_from_xpub(xpubd, compressed=False)

        xpub = (
            "xpub661MyMwAqRbcFtXgS5sYJABqqG9YLmC4Q1Rdap9gSE8NqtwybGhePY2g"
            "Z29ESFjqJoCu1Rupje8YtGqsefD265TMg7usUDFdp6W1EGMcet8"
        ).encode()
        self.assertRaises(ValueError, point_from_key, xpub, secp256r1)

        wif = b"KzyziFNa2m2WC84NDBG2ix3rQXYcKHndvCjTkmJQWuoadpQxmdmu"
        self.assertRaises(ValueError, point_from_key, wif, secp256r1)
        # point_from_key(wif, secp256r1)

        point_from_key(1, secp256r1)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

""" Base58 address functions.

Base58 encoding of public keys and scripts as addresses.
"""

from typing import Optional, Tuple

from .alias import Key, Octets, Script, String
from .base58 import b58decode, b58encode
from .bech32address import b32address_from_witness, witness_from_b32address
from .hashes import hash160_from_pubkey, hash160_from_script, hash256_from_script
from .network import _P2PKH_PREFIXES, _P2SH_PREFIXES, NETWORKS, network_from_key_value
from .scriptpubkey import payload_from_scriptPubKey, scriptPubKey_from_payload
from .utils import bytes_from_octets

# 1. Hash/WitnessProgram from pubkey/script
# imported from the hashes module

# 2. base58 address from HASH and vice versa


def b58address_from_h160(prefix: Octets, h160: Octets, network: str) -> bytes:
    "Encode a base58 address from the payload."

    prefix = bytes_from_octets(prefix)
    prefixes = NETWORKS[network]["p2pkh"], NETWORKS[network]["p2sh"]
    if prefix not in prefixes:
        raise ValueError(f"Invalid {network} base58 address prefix {prefix!r}")
    payload = prefix + bytes_from_octets(h160, 20)
    return b58encode(payload)


def h160_from_b58address(b58addr: String) -> Tuple[bytes, bytes, str, bool]:
    "Return the payload from a base58 address."

    if isinstance(b58addr, str):
        b58addr = b58addr.strip()

    payload = b58decode(b58addr, 21)
    prefix = payload[0:1]
    if prefix in _P2PKH_PREFIXES:
        network = network_from_key_value("p2pkh", prefix)
        is_script_hash = False
    elif prefix in _P2SH_PREFIXES:
        network = network_from_key_value("p2sh", prefix)
        is_script_hash = True
    else:
        raise ValueError(f"Invalid base58 address prefix {prefix!r}")

    return prefix, payload[1:], network, is_script_hash


# 1.+2. = 3. base58 address from pubkey/script


def p2pkh(
    key: Key, network: Optional[str] = None, compressed: Optional[bool] = None
) -> bytes:
    "Return the p2pkh base58 address corresponding to a public key."
    h160, network = hash160_from_pubkey(key, network, compressed)
    prefix = NETWORKS[network]["p2pkh"]
    return b58address_from_h160(prefix, h160, network)


def p2sh(script: Script, network: str = "mainnet") -> bytes:
    "Return the p2sh base58 address corresponding to a script."
    h160 = hash160_from_script(script)
    prefix = NETWORKS[network]["p2sh"]
    return b58address_from_h160(prefix, h160, network)


# 2b. base58 address from WitnessProgram
# it cannot be inverted because of the hash performed by p2sh


def b58address_from_witness(wp: Octets, network: str = "mainnet") -> bytes:
    "Encode a legacy base58 p2sh-wrapped SegWit address."

    length = len(wp)
    if length == 20:
        redeem_script = scriptPubKey_from_payload("p2wpkh", wp)
    elif length == 32:
        redeem_script = scriptPubKey_from_payload("p2wsh", wp)
    else:
        m = f"Invalid witness program length ({len(wp)})"
        raise ValueError(m)

    return p2sh(redeem_script, network)


# 1.+2b. = 3b. base58 (p2sh-wrapped) SegWit addresses from pubkey/script


def p2wpkh_p2sh(key: Key, network: Optional[str] = None) -> bytes:
    "Return the p2wpkh-p2sh base58 address corresponding to a pubkey."
    compressed = True  # needed to force check on pubkey
    witprog, network = hash160_from_pubkey(key, network, compressed)
    return b58address_from_witness(witprog, network)


def p2wsh_p2sh(wscript: Script, network: str = "mainnet") -> bytes:
    "Return the p2wsh-p2sh base58 address corresponding to a script."
    witprog = hash256_from_script(wscript)
    return b58address_from_witness(witprog, network)


##########################


def has_segwit_prefix(addr: String) -> bool:

    if isinstance(addr, str):
        str_addr = addr.strip()
        str_addr = str_addr.lower()
    else:
        str_addr = addr.decode("ascii")

    for net in NETWORKS:
        if str_addr.startswith(NETWORKS[net]["p2w"] + "1"):
            return True

    return False


def scriptPubKey_from_address(addr: String) -> Tuple[bytes, str]:
    "Return (scriptPubKey, network) from the input bech32/base58 address"

    if has_segwit_prefix(addr):
        # also check witness validity
        wv, wp, network, is_script_hash = witness_from_b32address(addr)
        if wv != 0:
            raise ValueError(f"Unmanaged witness version ({wv})")
        if is_script_hash:
            return scriptPubKey_from_payload("p2wsh", wp), network
        else:
            return scriptPubKey_from_payload("p2wpkh", wp), network
    else:
        _, h160, network, is_p2sh = h160_from_b58address(addr)
        if is_p2sh:
            return scriptPubKey_from_payload("p2sh", h160), network
        else:
            return scriptPubKey_from_payload("p2pkh", h160), network


def address_from_scriptPubKey(s: Script, network: str = "mainnet") -> bytes:
    "Return the bech32/base58 address from the input scriptPubKey."

    script_type, payload, m = payload_from_scriptPubKey(s)
    if script_type == "p2pk":
        raise ValueError("No address for p2pk script")
    if script_type == "p2ms" or isinstance(payload, list) or m != 0:
        raise ValueError("No address for p2ms script")
    if script_type == "nulldata":
        raise ValueError("No address for null data script")

    if script_type == "p2pkh":
        prefix = NETWORKS[network]["p2pkh"]
        return b58address_from_h160(prefix, payload, network)
    if script_type == "p2sh":
        prefix = NETWORKS[network]["p2sh"]
        return b58address_from_h160(prefix, payload, network)

    # 'p2wsh' or 'p2wpkh'
    return b32address_from_witness(0, payload, network)



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest
from os import path
import json

from btclib import bip32, slip32
from btclib import electrum


class TestMnemonicDictionaries(unittest.TestCase):
    def test_mnemonic(self):
        lang = "en"

        entropy = 0x110AAAA03974D093EDA670121023CD0772
        eversion = "standard"
        # FIXME: is the following mnemonic obtained in Electrum
        # from the above entropy?
        mnemonic = (
            "ability awful fetch liberty company spatial "
            "panda hat then canal ball crouch bunker"
        )
        mnemonic2 = electrum.mnemonic_from_entropy(entropy, eversion, lang)
        self.assertEqual(mnemonic, mnemonic2)

        entr = int(electrum.entropy_from_mnemonic(mnemonic, lang), 2)
        self.assertLess(entr - entropy, 0xFFF)

        # mnemonic version not in electrum allowed mnemonic versions
        eversion = "std"
        self.assertRaises(
            ValueError, electrum.mnemonic_from_entropy, entropy, eversion, lang
        )
        # electrum.mnemonic_from_entropy(entropy, eversion, lang)

        # unknown electrum mnemonic version (00c)
        unknown_version = (
            "ability awful fetch liberty company spatial "
            "panda hat then canal ball cross video"
        )
        self.assertRaises(
            ValueError, electrum.entropy_from_mnemonic, unknown_version, lang
        )
        # electrum.entropy_from_mnemonic(unknown_version, lang)

        passphrase = ""

        # unknown electrum mnemonic version (00c)
        self.assertRaises(
            ValueError, bip32.mxprv_from_electrum_mnemonic, unknown_version, passphrase
        )
        # bip32.mxprv_from_electrum_mnemonic(mnemonic, passphrase)

        xprv = (
            "xprv9s21ZrQH143K2tn5j4pmrLXkS6dkbuX6mFhJfCxAwN6ofRo5ddCrLRWo"
            "gKEs1AptPmLgrthKxU2csfBgkoKECWtj1XMRicRsoWawukaRQft"
        )
        xprv2 = bip32.mxprv_from_electrum_mnemonic(mnemonic, passphrase)
        self.assertEqual(xprv2, xprv.encode())

        eversion = "2fa"
        mnemonic = electrum.mnemonic_from_entropy(entropy, eversion, lang)
        # 2fa mnemonic version is not managed yet
        self.assertRaises(
            ValueError, bip32.mxprv_from_electrum_mnemonic, mnemonic, passphrase
        )
        # bip32.mxprv_from_electrum_mnemonic(mnemonic, passphrase)

        eversion = "2fa_segwit"
        mnemonic = electrum.mnemonic_from_entropy(entropy, eversion, lang)
        # 2fa_segwit mnemonic version is not managed yet
        self.assertRaises(
            ValueError, bip32.mxprv_from_electrum_mnemonic, mnemonic, passphrase
        )
        # bip32.mxprv_from_electrum_mnemonic(mnemonic, passphrase)

    def test_vectors(self):
        fname = "electrum_test_vectors.json"
        filename = path.join(path.dirname(__file__), "test_data", fname)
        with open(filename, "r") as f:
            test_vectors = json.load(f)

        lang = "en"
        for test_vector in test_vectors:
            mnemonic = test_vector[0]
            passphrase = test_vector[1]
            mxprv = test_vector[2]
            mxpub = test_vector[3]
            address = test_vector[4]  # "./0/0"

            if mnemonic != "":
                mxprv2 = bip32.mxprv_from_electrum_mnemonic(mnemonic, passphrase)
                self.assertEqual(mxprv2, mxprv.encode())

                eversion, mnemonic = electrum.version_from_mnemonic(mnemonic)
                entr = int(electrum.entropy_from_mnemonic(mnemonic, lang), 2)
                mnem = electrum.mnemonic_from_entropy(entr, eversion, lang)
                self.assertEqual(mnem, mnemonic)

            if mxprv != "":
                mxpub2 = bip32.xpub_from_xprv(mxprv)
                self.assertEqual(mxpub2, mxpub.encode())

            xpub = bip32.derive(mxpub, "./0/0")
            address2 = slip32.address_from_xpub(xpub).decode("ascii")
            self.assertEqual(address2, address)

        # version 2fa_segwit
        mnemonic = (
            "slender flight session office noodle  hand "
            "couple  option office  wait   uniform morning"
        )
        self.assertEqual("2fa_segwit", electrum.version_from_mnemonic(mnemonic)[0])

        # version 2fa
        mnemonic = (
            "history recycle company awful donor   fold "
            "beef    nominee hard    bleak bracket six"
        )
        self.assertEqual("2fa", electrum.version_from_mnemonic(mnemonic)[0])


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib import varint


class TestVarInt(unittest.TestCase):
    def test_conversion(self):

        i = 0xFC
        b = varint.encode(i)
        self.assertEqual(len(b), 1)
        self.assertEqual(varint.decode(b), i)

        i += 1
        b = varint.encode(i)
        self.assertEqual(len(b), 3)
        self.assertEqual(varint.decode(b), i)

        i = 0xFFFF
        b = varint.encode(i)
        self.assertEqual(len(b), 3)
        self.assertEqual(varint.decode(b), i)

        i += 1
        b = varint.encode(i)
        self.assertEqual(len(b), 5)
        self.assertEqual(varint.decode(b), i)

        i = 0xFFFFFFFF
        b = varint.encode(i)
        self.assertEqual(len(b), 5)
        self.assertEqual(varint.decode(b), i)

        i += 1
        b = varint.encode(i)
        self.assertEqual(len(b), 9)
        self.assertEqual(varint.decode(b), i)

        i = 0xFFFFFFFFFFFFFFFF
        b = varint.encode(i)
        self.assertEqual(len(b), 9)
        self.assertEqual(varint.decode(b), i)

        # integer too large (0x10000000000000000) for varint encoding
        i += 1
        self.assertRaises(ValueError, varint.encode, i)
        # varint.encode(i)

        self.assertEqual(varint.decode("6a"), 106)
        self.assertEqual(varint.decode("fd2602"), 550)
        self.assertEqual(varint.decode("fe703a0f00"), 998000)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest

from btclib.base58 import (
    _b58decode,
    _b58decode_to_int,
    _b58encode,
    _b58encode_from_int,
    b58decode,
    b58encode,
)


class TestBase58CheckEncoding(unittest.TestCase):
    def test_empty(self):
        self.assertEqual(_b58encode(b""), b"")
        self.assertEqual(_b58decode(_b58encode(b"")), b"")

    def test_hello_world(self):
        self.assertEqual(_b58encode(b"hello world"), b"StV1DL6CwTryKyV")
        self.assertEqual(_b58decode(b"StV1DL6CwTryKyV"), b"hello world")
        self.assertEqual(_b58decode(_b58encode(b"hello world")), b"hello world")
        self.assertEqual(_b58encode(_b58decode(b"StV1DL6CwTryKyV")), b"StV1DL6CwTryKyV")

    def test_trailing_zeros(self):
        self.assertEqual(_b58encode(b"\x00\x00hello world"), b"11StV1DL6CwTryKyV")
        self.assertEqual(_b58decode(b"11StV1DL6CwTryKyV"), b"\x00\x00hello world")
        self.assertEqual(
            _b58decode(_b58encode(b"\0\0hello world")), b"\x00\x00hello world"
        )
        self.assertEqual(
            _b58encode(_b58decode(b"11StV1DL6CwTryKyV")), b"11StV1DL6CwTryKyV"
        )

    def test_exceptions(self):
        # int is not hex-string or bytes
        self.assertRaises(TypeError, b58encode, 3)

        encoded = b58encode(b"test")

        # unexpected decoded length
        wrong_length = len(encoded) - 1
        self.assertRaises(ValueError, b58decode, encoded, wrong_length)

        # checksum is invalid
        invalidChecksum = encoded[:-4] + b"1111"
        self.assertRaises(ValueError, b58decode, invalidChecksum, 4)

        # non-ascii character
        self.assertRaises(ValueError, b58decode, "hèllo world")
        # b58decode("hèllo world")

    def test_wif(self):
        # https://en.bitcoin.it/wiki/Wallet_import_format
        prv = 0xC28FCA386C7A227600B2FE50B7CAE11EC86D3BF1FBE471BE89827E19D72AA1D

        uncompressedKey = b"\x80" + prv.to_bytes(32, byteorder="big")
        uncomprWIF = b"5HueCGU8rMjxEXxiPuD5BDku4MkFqeZyd4dZ1jvhTVqvbTLvyTJ"
        wif = b58encode(uncompressedKey)
        self.assertEqual(wif, uncomprWIF)
        key = b58decode(uncomprWIF)
        self.assertEqual(key, uncompressedKey)

        compressedKey = b"\x80" + prv.to_bytes(32, byteorder="big") + b"\x01"
        compressedWIF = b"KwdMAjGmerYanjeui5SHS7JkmpZvVipYvB2LJGU1ZxJwYvP98617"
        wif = b58encode(compressedKey)
        self.assertEqual(wif, compressedWIF)
        key = b58decode(compressedWIF)
        self.assertEqual(key, compressedKey)

        # string
        compressedWIF = b"KwdMAjGmerYanjeui5SHS7JkmpZvVipYvB2LJGU1ZxJwYvP98617"
        key = b58decode(compressedWIF)
        self.assertEqual(key, compressedKey)


def test_integers():
    digits = b"123456789ABCDEFGHJKLMNPQRSTUVWXYZabcdefghijkmnopqrstuvwxyz"
    for i in range(len(digits)):
        char = digits[i : i + 1]
        assert _b58decode_to_int(char) == i
        assert _b58encode_from_int(i) == char
    number = (
        "0111d38e5fc9071ffcd20b4a763cc9ae4f252bb4e4"
        "8fd66a835e252ada93ff480d6dd43dc62a641155a5"
    )
    n = int(number, 16)
    assert _b58decode_to_int(digits) == n
    assert _b58encode_from_int(n) == digits[1:]


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import secrets
import unittest

from btclib.utils import bytes_from_octets, hash160, hash256, int_from_integer

int_with_whitespaces = (
    " 0C 28FCA386C7A227600B2FE50B7CAE11EC86D3BF1FBE471BE89827E19D72AA1D  "
)


class TestUtils(unittest.TestCase):
    def test_utils(self):
        b = bytes_from_octets(int_with_whitespaces)
        s = b.hex()  # lower case, no spaces
        self.assertNotEqual(int_with_whitespaces, s)
        self.assertEqual(hash160(int_with_whitespaces), hash160(s))
        self.assertEqual(hash256(int_with_whitespaces), hash256(s))

        i = secrets.randbits(256)
        self.assertEqual(i, int_from_integer(i))
        self.assertEqual(i, int_from_integer(i.to_bytes(32, "big")))
        self.assertEqual(i, int_from_integer(hex(i)))


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest
from hashlib import sha256

from btclib import dsa, ssa
from btclib.signtocontract import ecdsa_commit_sign, ecssa_commit_sign, verify_commit
from btclib.curves import secp256k1 as ec


class TestSignToContract(unittest.TestCase):
    def test_signtocontract(self):
        m = b"to be signed"
        c = b"to be committed"

        prv, pub = dsa.gen_keys()
        dsa_sig, dsa_receipt = ecdsa_commit_sign(c, m, prv, None)
        self.assertIsNone(dsa._verify(m, pub, dsa_sig, ec, sha256))
        self.assertTrue(verify_commit(c, dsa_receipt))

        # 32 bytes message for ECSSA
        m = sha256(m).digest()
        prv, pub = ssa.gen_keys()
        ssa_sig, ssa_receipt = ecssa_commit_sign(c, m, prv, None)
        self.assertIsNone(ssa._verify(m, pub, ssa_sig, ec, sha256))
        self.assertTrue(verify_commit(c, ssa_receipt))


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic Curve Schnorr Signature Algorithm (ECSSA).

This implementation is according to BIP340-Schnorr:

https://github.com/bitcoin/bips/blob/master/bip-0340.mediawiki

Differently from ECDSA, the BIP340-Schnorr scheme supports
messages of size hsize only.

It also uses as public key the x-coordinate (field element)
of the curve point associated to the private key 0 < q < n.
Therefore, for sepcp256k1 the public key size is 32 bytes.
Arguably, the knowledge of q as the discrete logarithm of Q
also implies the knowledge of n-q as discrete logarithm of -Q.
As such, {q, n-q} can be considered a single private key and
{Q, -Q} the associated public key characterized by the shared x_Q.

Also, BIP340 advocates its own SHA256 modification as hash function:
TaggedHash(tag, x) = SHA256(SHA256(tag)||SHA256(tag)||x)
The rationale is to make BIP340 signatures invalid for anything else
but Bitcoin and vice versa.

TaggedHash is used for both the challenge (with tag 'BIPSchnorr')
and the deterministic nonce (with tag 'BIPSchnorrDerive').

To allow for secure batch verification of multiple signatures,
BIP340-Schnorr uses a challenge that prevents public key recovery
from signature: c = TaggedHash('BIPSchnorr', x_k||x_Q||msg).

A custom deterministic algorithm for the ephemeral key (nonce)
is used for signing, instead of the RFC6979 standard:
k = TaggedHash('BIPSchnorrDerive', q||msg)

Finally, BIP340-Schnorr adopts a robust [r][s] custom serialization
format, instead of the loosely specified ASN.1 DER standard.
The signature size is p-size*n-size, where p-size is the field element
(curve point coordinate) byte size and n-size is the scalar
(curve point multiplication coefficient) byte size.
For sepcp256k1 the resulting signature size is 64 bytes.
"""

import secrets
from hashlib import sha256
from typing import List, Sequence, Tuple, Union

from .alias import HashF, JacPoint, Octets, Point, PrvKey, SSASig, SSASigTuple
from .bip32 import BIP32KeyDict
from .curve import Curve
from .curvemult import _double_mult, _mult_jac, _multi_mult
from .curves import secp256k1
from .numbertheory import mod_inv
from .to_prvkey import int_from_prvkey
from .to_pubkey import point_from_pubkey
from .utils import bytes_from_octets, int_from_bits

# TODO relax the p_ThreeModFour requirement


BIP340PubKey = Union[int, bytes, str, BIP32KeyDict]


def _validate_sig(r: int, s: int, ec: Curve) -> None:

    # BIP340 is only defined for curves whose field prime p = 3 % 4
    ec.require_p_ThreeModFour()

    # Fail if r is not a field element, i.e. not a valid x-coordinate
    ec.y(r)

    # Fail if s is not [0, n-1].
    if not 0 <= s < ec.n:
        raise ValueError(f"s ({hex(s)}) not in [0, n-1]")


def deserialize(sig: SSASig, ec: Curve = secp256k1) -> SSASigTuple:
    """Return the verified components of the provided BIP340 signature.

    The BIP340 signature can be represented as (r, s) tuple
    or as binary [r][s] compact representation.
    """

    if isinstance(sig, tuple):
        r, s = sig
    else:
        sig = bytes_from_octets(sig, ec.psize + ec.nsize)
        r = int.from_bytes(sig[: ec.psize], byteorder="big")
        s = int.from_bytes(sig[ec.nsize :], byteorder="big")

    _validate_sig(r, s, ec)
    return r, s


def serialize(x_K: int, s: int, ec: Curve = secp256k1) -> bytes:
    "Return the BIP340 signature as [r][s] compact representation."

    _validate_sig(x_K, s, ec)
    sig = x_K.to_bytes(ec.psize, "big") + s.to_bytes(ec.nsize, "big")
    return sig


def gen_keys(prvkey: PrvKey = None, ec: Curve = secp256k1) -> Tuple[int, int]:
    "Return a BIP340 private/public (int, int) key-pair."
    # BIP340 is only defined for curves whose field prime p = 3 % 4
    ec.require_p_ThreeModFour()

    if prvkey is None:
        # q in the range [1, ec.n-1]
        q = 1 + secrets.randbelow(ec.n - 1)
    else:
        q = int_from_prvkey(prvkey, ec)

    QJ = _mult_jac(q, ec.GJ, ec)
    x_Q = ec._x_aff_from_jac(QJ)
    return q, x_Q


# This implementation can be sped up by storing the midstate after hashing
# tag_hash instead of rehashing it all the time.
def _tagged_hash(tag: str, m: bytes, hf: HashF) -> bytes:
    t = tag.encode()
    h1 = hf()
    h1.update(t)
    tag_hash = h1.digest()
    h2 = hf()
    h2.update(tag_hash + tag_hash + m)
    return h2.digest()


def _k(m: bytes, q: int, ec: Curve, hf: HashF) -> int:

    # assume the random oracle model for the hash function,
    # i.e. hash values can be considered uniformly random

    # Note that in general, taking a uniformly random integer
    # modulo the curve order n would produce a biased result.
    # However, if the order n is sufficiently close to 2^hlen,
    # then the bias is not observable:
    # e.g. for secp256k1 and sha256 1-n/2^256 it is about 1.27*2^-128

    # the unbiased implementation is provided here,
    # which works also for very-low-cardinality test curves
    t = q.to_bytes(ec.nsize, "big") + m
    while True:
        t = _tagged_hash("BIPSchnorrDerive", t, hf)
        # The following lines would introduce a bias
        # k = int.from_bytes(t, 'big') % ec.n
        # k = int_from_bits(t, ec.nlen) % ec.n
        k = int_from_bits(t, ec.nlen)  # candidate k
        if 0 < k < ec.n:  # acceptable value for k
            return k  # successful candidate


def k(m: Octets, prv: PrvKey, ec: Curve = secp256k1, hf: HashF = sha256) -> int:
    """Return a BIP340 deterministic ephemeral key (nonce)."""

    # The message m: a hlen array
    m = bytes_from_octets(m, hf().digest_size)

    # The secret key d: an integer in the range 1..n-1.
    q = int_from_prvkey(prv, ec)

    return _k(m, q, ec, hf)


def _challenge(r: int, x_Q: int, m: bytes, ec: Curve, hf: HashF) -> int:

    # note that only x_Q is needed
    # if Q is Jacobian y_Q calculation can be avoided

    t = r.to_bytes(ec.psize, "big")
    t += x_Q.to_bytes(ec.psize, "big")
    # m size must have been already checked to be equal to hsize
    t += m
    t = _tagged_hash("BIPSchnorr", t, hf)
    c = int_from_bits(t, ec.nlen) % ec.n
    if c == 0:
        raise ValueError("Invalid (zero) challenge")
    return c


def sign(
    m: Octets,
    prvkey: PrvKey,
    k: PrvKey = None,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> SSASigTuple:
    """Sign message according to BIP340 signature algorithm."""

    # BIP340 is only defined for curves whose field prime p = 3 % 4
    ec.require_p_ThreeModFour()

    # The message m: a hlen array
    m = bytes_from_octets(m, hf().digest_size)

    # The secret key q: an integer in the range 1..n-1.
    q = int_from_prvkey(prvkey, ec)
    QJ = _mult_jac(q, ec.GJ, ec)
    x_Q = ec._x_aff_from_jac(QJ)
    if not ec.has_square_y(QJ):
        q = ec.n - q

    # The nonce k: an integer in the range 1..n-1.
    if k is None:
        k = _k(m, q, ec, hf)
    else:
        k = int_from_prvkey(k, ec)
    KJ = _mult_jac(k, ec.GJ, ec)
    x_K = ec._x_aff_from_jac(KJ)
    if not ec.has_square_y(KJ):
        k = ec.n - k

    # Let c = int(hf(bytes(x_K) || bytes(x_Q) || m)) mod n.
    c = _challenge(x_K, x_Q, m, ec, hf)

    # s=0 is ok: in verification there is no inverse of s
    s = (k + c * q) % ec.n

    return x_K, s


def _to_bip340_point(x_Q: BIP340PubKey, ec: Curve = secp256k1) -> Point:
    """Return a verified-as-valid BIP340 public key as Point tuple.

    It supports:

    - BIP32 extended keys (bytes, string, or BIP32KeyDict)
    - SEC Octets (bytes or hex-string, with 02, 03, or 04 prefix)
    - BIP340 Octets (bytes or hex-string, p-size Point x-coordinate)
    - native tuple
    """

    # BIP 340 key as integer
    if isinstance(x_Q, int):
        y_Q = ec.y_quadratic_residue(x_Q, True)
        return x_Q, y_Q
    else:
        try:
            x_Q = point_from_pubkey(x_Q, ec)[0]
            y_Q = ec.y_quadratic_residue(x_Q, True)
            return x_Q, y_Q
        except Exception:
            pass

    # BIP 340 key as bytes or hex-string
    if isinstance(x_Q, str) or isinstance(x_Q, bytes):
        Q = bytes_from_octets(x_Q, ec.psize)
        x_Q = int.from_bytes(Q, "big")
        y_Q = ec.y_quadratic_residue(x_Q, True)
    else:
        raise ValueError("not a BIP340 public key")

    return x_Q, y_Q


def _to_sig(sig: SSASig, ec: Curve) -> SSASigTuple:
    if isinstance(sig, tuple):
        r, s = sig
        _validate_sig(r, s, ec)
    else:
        # it is a serialized signature
        r, s = deserialize(sig, ec)
    return r, s


def _verify(m: Octets, Q: BIP340PubKey, sig: SSASig, ec: Curve, hf: HashF) -> None:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    # BIP340 is only defined for curves whose field prime p = 3 % 4
    ec.require_p_ThreeModFour()

    # The message m: a hlen array
    m = bytes_from_octets(m, hf().digest_size)

    r, s = _to_sig(sig, ec)

    x_Q, y_Q = _to_bip340_point(Q, ec)
    QJ = x_Q, y_Q, 1

    # Let c = int(hf(bytes(r) || bytes(Q) || m)) mod n.
    c = _challenge(r, x_Q, m, ec, hf)

    # Let K = sG - eQ.
    # in Jacobian coordinates
    KJ = _double_mult(ec.n - c, QJ, s, ec.GJ, ec)

    # Fail if infinite(KJ).
    # Fail if jacobi(y_K) ≠ 1.
    assert ec.has_square_y(KJ), "y_K is not a quadratic residue"

    # Fail if x_K ≠ r
    assert KJ[0] == KJ[2] * KJ[2] * r % ec.p, "Signature verification failed"


def verify(
    m: Octets, Q: BIP340PubKey, sig: SSASig, ec: Curve = secp256k1, hf: HashF = sha256
) -> bool:
    """Verify the BIP340 signature of the provided message."""

    # try/except wrapper for the Errors raised by _verify
    try:
        _verify(m, Q, sig, ec, hf)
    except Exception:
        return False
    else:
        return True


def _recover_pubkeys(c: int, r: int, s: int, ec: Curve) -> int:
    # Private function provided for testing purposes only.

    KJ = r, ec.y_quadratic_residue(r, True), 1

    e1 = mod_inv(c, ec.n)
    QJ = _double_mult(ec.n - e1, KJ, e1 * s, ec.GJ, ec)
    assert QJ[2] != 0, "how did you do that?!?"
    return ec._x_aff_from_jac(QJ)


def crack_prvkey(
    m1: Octets,
    sig1: SSASig,
    m2: Octets,
    sig2: SSASig,
    Q: BIP340PubKey,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> Tuple[int, int]:

    m1 = bytes_from_octets(m1, hf().digest_size)
    r1, s1 = _to_sig(sig1, ec)
    m2 = bytes_from_octets(m2, hf().digest_size)
    r2, s2 = _to_sig(sig2, ec)
    x_Q = _to_bip340_point(Q, ec)[0]

    if r1 != r2:
        raise ValueError("Not the same r in signatures")
    if s1 == s2:
        raise ValueError("Identical signatures")

    c1 = _challenge(r1, x_Q, m1, ec, hf)
    c2 = _challenge(r2, x_Q, m2, ec, hf)
    q = (s1 - s2) * mod_inv(c2 - c1, ec.n) % ec.n
    k = (s1 + c1 * q) % ec.n
    return q, k


def _batch_verify(
    ms: Sequence[Octets],
    Qs: Sequence[BIP340PubKey],
    sigs: Sequence[SSASig],
    ec: Curve,
    hf: HashF,
) -> None:

    batch_size = len(Qs)
    if len(ms) != batch_size:
        errMsg = f"mismatch between number of pubkeys ({batch_size}) "
        errMsg += f"and number of messages ({len(ms)})"
        raise ValueError(errMsg)
    if len(sigs) != batch_size:
        errMsg = f"mismatch between number of pubkeys ({batch_size}) "
        errMsg += f"and number of signatures ({len(sigs)})"
        raise ValueError(errMsg)

    if batch_size < 2:
        return _verify(ms[0], Qs[0], sigs[0], ec, hf)

    # BIP340 is only defined for curves whose field prime p = 3 % 4
    ec.require_p_ThreeModFour()

    t = 0
    scalars: List[int] = list()
    points: List[JacPoint] = list()
    for i, (m, Q, sig) in enumerate(zip(ms, Qs, sigs)):
        m = bytes_from_octets(m, hf().digest_size)

        r, s = _to_sig(sig, ec)
        KJ = r, ec.y_quadratic_residue(r, True), 1

        x_Q, y_Q = _to_bip340_point(Q, ec)
        QJ = x_Q, y_Q, 1

        c = _challenge(r, x_Q, m, ec, hf)

        # a in [1, n-1]
        # deterministically generated using a CSPRNG seeded by a
        # cryptographic hash (e.g., SHA256) of all inputs of the
        # algorithm, or randomly generated independently for each
        # run of the batch verification algorithm
        a = 1 if i == 0 else 1 + secrets.randbelow(ec.n - 1)
        scalars.append(a)
        points.append(KJ)
        scalars.append(a * c % ec.n)
        points.append(QJ)
        t += a * s

    TJ = _mult_jac(t, ec.GJ, ec)
    RHSJ = _multi_mult(scalars, points, ec)

    # return T == RHS, checked in Jacobian coordinates
    RHSZ2 = RHSJ[2] * RHSJ[2]
    TZ2 = TJ[2] * TJ[2]
    precondition = TJ[0] * RHSZ2 % ec.p == RHSJ[0] * TZ2 % ec.p
    assert precondition, "Signature verification precondition failed"

    valid_sig = TJ[1] * RHSZ2 * RHSJ[2] % ec.p == RHSJ[1] * TZ2 * TJ[2] % ec.p
    assert valid_sig, "Signature verification failed"


def batch_verify(
    m: Sequence[Octets],
    Q: Sequence[BIP340PubKey],
    sig: Sequence[SSASig],
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> bool:
    """Batch verification of BIP340 signatures."""

    # try/except wrapper for the Errors raised by _batch_verify
    try:
        _batch_verify(m, Q, sig, ec, hf)
    except Exception:
        return False
    else:
        return True



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Aliases

mypy aliases, documenting also coding imput conventions.
"""


from typing import Any, Callable, Iterable, List, Tuple, TypedDict, Union

# binary octets are eight-bit bytes or hex-string (not text string)
#
# use bytes_from_octets to properly convert to bytes
#
# used for script, h160 (20 bytes), h256 (32 bytes),
# bip32version (4 bytes), sighash (1 byte),
# dersig (DER serialization of ECDSA signature),
# msgsig (Bitcoin message compact signature serialization, 65 bytes),
# etc.
Octets = Union[bytes, str]

# hex-string or bytes representation of an int
Integer = Union[bytes, str, int]

# bytes or text string (not hex-string)
#
# to convert to bytes, just encode()
# e.g. in order to sign a message
#    if isinstance(msg, str):
#        msg = msg.encode()
#
# in many cases (e.g. b58addr, b32addr, wif, bip32key)
# leading/trailing blanks can be stripped
# if isinstance(b58addr, str):
#     b58addr = b58addr.strip()
#
# in those cases often there is no need to encode() to bytes
# as b58decode/b32decode/etc. will take care of that
# also those string are 'ascii', a subset of 'utf-8'
String = Union[bytes, str]


# Hash digest constructor: it may be any name suitable to hashlib.new()
HashF = Callable[[], Any]
# HashF = Callable[[Any], Any]


# Elliptic curve point in affine coordinates.
# Warning: to make Point a NamedTuple would slow down the code
Point = Tuple[int, int]

# Note that the infinity point in affine coordinates is INF = (int, 0)
# (no affine point has y=0 coordinate in a group of prime order n).
# It can be checked with 'INF[1] == 0'
# The x-coordinate is arbitrary: 7 is preferred
# because it is not a field element in secp256k1
INF = 7, 0


# Elliptic curve point in Jacobian coordinates.
JacPoint = Tuple[int, int, int]

# Infinity point in Jacobian coordinates is INF = (int, int, 0).
# It can be checked with 'INF[2] == 0'
# The default x and y coordinates are arbitrary:
# 7, 0 are used because those are what one would obtain
# from the generic affine to jacobian transformation
# of the INF Point
# QJ = Q[0], Q[1], 1 if Q[1] else 0
INFJ = 7, 0, 0


# the main internal representation of entropy is binary 0/1 string
BinStr = str
# but int or bytes are fine too
Entropy = Union[BinStr, int, bytes]


# BIP 32 derivation path
# absolute path as "m/44h/0'/1H/0/10" string
# relative path as "./0/10" string
# relative path as sequence of integer indexes
# relative one level child derivation with single 4-bytes index
# relative one level child derivation with single integer index
# TODO: allow also Iterable[bytes], while making mypy happy
Path = Union[str, Iterable[int], int, bytes]


# BIP 32 extended key as a TypedDict
class BIP32KeyDict(TypedDict):
    version: bytes
    depth: int
    parent_fingerprint: bytes
    index: bytes
    chain_code: bytes
    key: bytes


BIP32Key = Union[BIP32KeyDict, String]

# private key inputs:
# integer as Union[int, Octets]
# BIP32key as BIP32Key
# WIF as String
#
# BIP32key and WIF also provide extra info about
# network and (un)compressed-pubkey-derivation
PrvKey = Union[int, bytes, str, BIP32KeyDict]

# public key inputs:
# elliptic curve point as Union[Octets, BIP32Key, Point]
PubKey = Union[bytes, str, BIP32KeyDict, Point]

# public or private key input,
# usable wherever a PubKey is logically expected
Key = Union[int, bytes, str, BIP32KeyDict, Point]

# ECDSA signature
# (r, s)
# both r and s are scalar: 0 < r < ec.n, 0 < s < ec.n
DSASigTuple = Tuple[int, int]
# DSASigTuple or DER serialization (bytes or hex-string, no sighash)
DSASig = Union[DSASigTuple, Octets]


# Bitcoin message signature
# (rf, r, s), where r and s are the components of a DSASigTuple
BMSigTuple = Tuple[int, int, int]
# BMSigTuple or base64 65-bytes serialization (bytes or hex-string)
BMSig = Union[BMSigTuple, Octets]


# BIP340-Schnorr signature
# (r, s)
# r is a _field_element_, 0 <= r < ec.p
# s is a scalar, 0 <= s < ec.n (yes, for BIP340-Schnorr it can be zero)
# (p is the field prime, n is the curve order)
SSASigTuple = Tuple[int, int]
# SSASigTuple or BIP340-Schnorr 64-bytes serialization (bytes or hex-string)
SSASig = Union[SSASigTuple, Octets]

# the integers [0-16] are shorcuts for 'OP_0'-'OP_16'
# the integer -1 is a shorcut for 'OP_1NEGATE'
# other integers are bytes encoded (require push operation)
# ascii str are for opcodes (e.g. 'OP_HASH160')
# Octets are for data to be pushed
Token = Union[int, str, bytes]

# Bitcoin script expressed as List[Token]
# e.g. [OP_HASH160, script_h160, OP_EQUAL]
# or Octets of its byte-encoded representation
Script = Union[Octets, List[Token]]



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

"""Elliptic Curve Digital Signature Algorithm (ECDSA).

   Implementation according to SEC 1 v.2:

   http://www.secg.org/sec1-v2.pdf

   specialized with bitcoin canonical 'low-s' encoding.
"""

import secrets
from hashlib import sha256
from typing import List, Optional, Tuple

from . import der
from .alias import DSASig, DSASigTuple, HashF, JacPoint, Key, Point, PrvKey, String
from .curve import Curve
from .curvemult import _double_mult, _mult_jac
from .curves import secp256k1
from .numbertheory import mod_inv
from .rfc6979 import _rfc6979
from .to_prvkey import int_from_prvkey
from .to_pubkey import point_from_key
from .utils import int_from_bits

# _validate_sig, deserialize and serialize are basically just wrappers
# for the equivalent functions in the der module


def _validate_sig(r: int, s: int, ec: Curve = secp256k1) -> None:
    return der._validate_sig(r, s, None, ec)


def deserialize(sig: DSASig, ec: Curve = secp256k1) -> DSASigTuple:
    """Return the verified components of the provided ECDSA signature.

    The ECDSA signature can be represented as (r, s) tuple or
    as strict ASN.1 DER binary representation.
    """

    if isinstance(sig, tuple):
        r, s = sig
        _validate_sig(r, s, ec)
        return r, s
    else:
        return der._deserialize(sig, ec)[0:2]


def serialize(r: int, s: int, ec: Curve = secp256k1) -> bytes:
    "Return the ECDSA signature as strict ASN.1 DER representation."

    return der._serialize(r, s, None, ec)


def gen_keys(prvkey: PrvKey = None, ec: Curve = secp256k1) -> Tuple[int, Point]:
    "Return a private/public (int, Point) key-pair."

    if prvkey is None:
        # q in the range [1, ec.n-1]
        q = 1 + secrets.randbelow(ec.n - 1)
    else:
        q = int_from_prvkey(prvkey, ec)

    QJ = _mult_jac(q, ec.GJ, ec)
    Q = ec._aff_from_jac(QJ)
    # q.to_bytes(ec.nsize, 'big')
    # bytes_from_point(Q, ec, compressed)
    return q, Q


def _challenge(msg: String, ec: Curve, hf: HashF) -> int:

    if isinstance(msg, str):
        msg = msg.encode()

    # Steps numbering follows SEC 1 v.2 section 4.1.3
    h = hf()
    h.update(msg)
    mhd = h.digest()  # 4
    # leftmost ec.nlen bits %= ec.n
    c = int_from_bits(mhd, ec.nlen) % ec.n  # 5
    return c


def _sign(c: int, q: int, k: int, ec: Curve) -> DSASigTuple:
    # Private function for testing purposes: it allows to explore all
    # possible value of the challenge c (for low-cardinality curves).
    # It assume that c is in [0, n-1], while q and k are in [1, n-1]

    # Steps numbering follows SEC 1 v.2 section 4.1.3

    KJ = _mult_jac(k, ec.GJ, ec)  # 1

    # affine x-coordinate of K (field element)
    K_x = (KJ[0] * mod_inv(KJ[2] * KJ[2], ec.p)) % ec.p
    # mod n makes it a scalar
    r = K_x % ec.n  # 2, 3
    if r == 0:  # r≠0 required as it multiplies the public key
        raise ValueError("r = 0, failed to sign")

    s = mod_inv(k, ec.n) * (c + r * q) % ec.n  # 6
    if s == 0:  # s≠0 required as verify will need the inverse of s
        raise ValueError("s = 0, failed to sign")

    # bitcoin canonical 'low-s' encoding for ECDSA signatures
    # it removes signature malleability as cause of transaction malleability
    # see https://github.com/bitcoin/bitcoin/pull/6769
    if s > ec.n / 2:
        s = ec.n - s  # s = - s % ec.n

    return r, s


# RFC6979 is used for deterministic nonce
def sign(
    msg: String,
    prvkey: PrvKey,
    k: Optional[PrvKey] = None,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> DSASigTuple:
    """ECDSA signature with canonical low-s encoding.

    Implemented according to SEC 1 v.2
    The message m is first processed by hf, yielding the value

        mhd = hf(m),

    a sequence of bits of length *hlen*.

    Normally, hf is chosen such that its output length *hlen* is
    roughly equal to *nlen*, the bit-length of the group order *n*,
    since the overall security of the signature scheme will depend on
    the smallest of *hlen* and *nlen*; however, the ECDSA standard
    supports all combinations of *hlen* and *nlen*.

    See https://tools.ietf.org/html/rfc6979#section-3.2
    """

    c = _challenge(msg, ec, hf)  # 4, 5

    # The secret key q: an integer in the range 1..n-1.
    # SEC 1 v.2 section 3.2.1
    q = int_from_prvkey(prvkey, ec)

    if k is None:
        k = _rfc6979(c, q, ec, hf)  # 1
    else:
        k = int_from_prvkey(k, ec)

    # second part delegated to helper function
    return _sign(c, q, k, ec)


def _to_sig(sig: DSASig, ec: Curve) -> DSASigTuple:
    if isinstance(sig, tuple):
        r, s = sig
        _validate_sig(r, s, ec)
    else:
        # it is a DER serialized signature
        # sighash is not needed
        r, s = deserialize(sig, ec)
    return r, s


def _verhlp(c: int, QJ: JacPoint, r: int, s: int, ec: Curve) -> None:
    # Private function for test/dev purposes

    w = mod_inv(s, ec.n)
    u = c * w % ec.n
    v = r * w % ec.n  # 4
    # Let K = u*G + v*Q.
    KJ = _double_mult(v, QJ, u, ec.GJ, ec)  # 5

    # Fail if infinite(K).
    assert KJ[2] != 0, "how did you do that?!?"  # 5

    # affine x-coordinate of K
    K_x = (KJ[0] * mod_inv(KJ[2] * KJ[2], ec.p)) % ec.p
    x = K_x % ec.n  # 6, 7
    # Fail if r ≠ K_x %n.
    assert r == x, "Signature verification failed"  # 8


def _verify(msg: String, P: Key, sig: DSASig, ec: Curve, hf: HashF) -> None:
    # Private function for test/dev purposes
    # It raises Errors, while verify should always return True or False

    r, s = _to_sig(sig, ec)  # 1

    c = _challenge(msg, ec, hf)  # 2, 3

    Q = point_from_key(P, ec)
    QJ = Q[0], Q[1], 1 if Q[1] else 0

    # second part delegated to helper function
    _verhlp(c, QJ, r, s, ec)


def verify(
    msg: String, P: Key, sig: DSASig, ec: Curve = secp256k1, hf: HashF = sha256
) -> bool:
    """ECDSA signature verification (SEC 1 v.2 section 4.1.4)."""

    # try/except wrapper for the Errors raised by _verify
    try:
        _verify(msg, P, sig, ec, hf)
    except Exception:
        return False
    else:
        return True


def recover_pubkeys(
    msg: String, sig: DSASig, ec: Curve = secp256k1, hf: HashF = sha256
) -> List[Point]:
    """ECDSA public key recovery (SEC 1 v.2 section 4.1.6).

    See also:
    https://crypto.stackexchange.com/questions/18105/how-does-recovering-the-public-key-from-an-ecdsa-signature-work/18106#18106
    """

    c = _challenge(msg, ec, hf)  # 1.5

    r, s = _to_sig(sig, ec)

    QJs = _recover_pubkeys(c, r, s, ec)
    return [ec._aff_from_jac(QJ) for QJ in QJs]


# TODO: use _recover_pubkey to avoid code duplication
def _recover_pubkeys(c: int, r: int, s: int, ec: Curve) -> List[JacPoint]:
    # Private function provided for testing purposes only.

    # precomputations
    r1 = mod_inv(r, ec.n)
    r1s = r1 * s % ec.n
    r1e = -r1 * c % ec.n
    keys: List[JacPoint] = list()
    # r = K[0] % ec.n
    # if ec.n < K[0] < ec.p (likely when cofactor ec.h > 1)
    # then both x=r and x=r+ec.n must be tested
    for j in range(ec.h):  # 1
        # affine x-coordinate of K (field element)
        x = (r + j * ec.n) % ec.p  # 1.1
        # two possible y-coordinates, i.e. two possible keys for each cycle
        try:
            # even root first for bitcoin message signing compatibility
            yodd = ec.y_odd(x, False)
            KJ = x, yodd, 1  # 1.2, 1.3, and 1.4
            # 1.5 has been performed in the recover_pubkeys calling function
            Q1J = _double_mult(r1s, KJ, r1e, ec.GJ, ec)  # 1.6.1
            try:
                _verhlp(c, Q1J, r, s, ec)  # 1.6.2
            except Exception:
                pass
            else:
                keys.append(Q1J)  # 1.6.2
            KJ = x, ec.p - yodd, 1  # 1.6.3
            Q2J = _double_mult(r1s, KJ, r1e, ec.GJ, ec)
            try:
                _verhlp(c, Q2J, r, s, ec)  # 1.6.2
            except Exception:
                pass
            else:
                keys.append(Q2J)  # 1.6.2
        except Exception:  # K is not a curve point
            pass
    return keys


def _recover_pubkey(key_id: int, c: int, r: int, s: int, ec: Curve) -> JacPoint:
    # Private function provided for testing purposes only.

    # precomputations
    r1 = mod_inv(r, ec.n)
    r1s = r1 * s % ec.n
    r1e = -r1 * c % ec.n
    # r = K[0] % ec.n
    # if ec.n < K[0] < ec.p (likely when cofactor ec.h > 1)
    # then both x=r and x=r+ec.n must be tested
    j = key_id & 0b110  # allow for key_id in [0, 7]
    x = (r + j * ec.n) % ec.p  # 1.1

    # even root first for Bitcoin Core compatibility
    i = key_id & 0b01
    y = ec.y_odd(x, i)
    KJ = x, y, 1  # 1.2, 1.3, and 1.4
    # 1.5 has been performed in the recover_pubkeys calling function
    QJ = _double_mult(r1s, KJ, r1e, ec.GJ, ec)  # 1.6.1
    _verhlp(c, QJ, r, s, ec)  # 1.6.2
    return QJ


def crack_prvkey(
    m1: String,
    sig1: DSASig,
    m2: String,
    sig2: DSASig,
    ec: Curve = secp256k1,
    hf: HashF = sha256,
) -> Tuple[int, int]:

    r1, s1 = _to_sig(sig1, ec)
    r2, s2 = _to_sig(sig2, ec)
    if r1 != r2:
        raise ValueError("Not the same r in signatures")
    if s1 == s2:
        raise ValueError("Identical signatures")

    c1 = _challenge(m1, ec, hf)
    c2 = _challenge(m2, ec, hf)
    k = (c1 - c2) * mod_inv(s1 - s2, ec.n) % ec.n
    q = (s2 * k - c2) * mod_inv(r1, ec.n) % ec.n
    return q, k



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

import unittest
from hashlib import sha1
from hashlib import sha256 as hf

from btclib import dsa
from btclib.alias import INF
from btclib.curvemult import _mult_jac, double_mult, mult
from btclib.curves import CURVES, secp256k1
from btclib.numbertheory import mod_inv
from btclib.secpoint import bytes_from_point, point_from_octets
from btclib.tests.test_curves import low_card_curves

secp112r2 = CURVES["secp112r2"]
secp160r1 = CURVES["secp160r1"]


class TestDSA(unittest.TestCase):
    def test_signature(self):

        ec = secp256k1
        q, Q = dsa.gen_keys(0x1)
        msg = "Satoshi Nakamoto"
        sig = dsa.sign(msg, q)
        self.assertEqual(sig, dsa.deserialize(sig))
        # https://bitcointalk.org/index.php?topic=285142.40
        # Deterministic Usage of DSA and ECDSA (RFC 6979)
        exp_sig = (
            0x934B1EA10A4B3C1757E2B0C017D0B6143CE3C9A7E6A4A49860D7A6AB210EE3D8,
            0x2442CE9D2B916064108014783E923EC36B49743E2FFA1C4496F01A512AAFD9E5,
        )
        r, s = sig
        self.assertEqual(sig[0], exp_sig[0])
        self.assertIn(sig[1], (exp_sig[1], secp256k1.n - exp_sig[1]))

        self.assertTrue(dsa.verify(msg, Q, sig))

        # malleability
        malleated_sig = (r, secp256k1.n - s)
        self.assertTrue(dsa.verify(msg, Q, malleated_sig))

        keys = dsa.recover_pubkeys(msg, sig)
        self.assertTrue(len(keys) == 2)
        self.assertIn(Q, keys)

        fmsg = "Craig Wright"
        self.assertFalse(dsa.verify(fmsg, Q, sig))

        fdsasig = (sig[0], sig[1], sig[1])
        self.assertFalse(dsa.verify(msg, Q, fdsasig))
        self.assertRaises(ValueError, dsa._verify, msg, Q, fdsasig, ec, hf)

        _, fQ = dsa.gen_keys()
        self.assertFalse(dsa.verify(msg, fQ, sig))

        # r not in [1, n-1]
        invalid_dassig = 0, sig[1]
        self.assertFalse(dsa.verify(msg, Q, invalid_dassig))

        # s not in [1, n-1]
        invalid_dassig = sig[0], 0
        self.assertFalse(dsa.verify(msg, Q, invalid_dassig))

        # pubkey = INF
        self.assertRaises(ValueError, dsa._verify, msg, INF, sig, ec, hf)
        # dsa._verify(msg, INF, sig, ec, hf)

        # private key not in [1, n-1]
        self.assertRaises(ValueError, dsa.sign, msg, 0)
        # dsa.sign(msg, 0)

        # ephemeral key not in [1, n-1]
        self.assertRaises(ValueError, dsa.sign, msg, 1, 0)
        # dsa.sign(msg, 1, 0)

    def test_gec(self):
        """GEC 2: Test Vectors for SEC 1, section 2

            http://read.pudn.com/downloads168/doc/772358/TestVectorsforSEC%201-gec2.pdf
        """
        # 2.1.1 Scheme setup
        ec = secp160r1
        hf = sha1

        # 2.1.2 Key Deployment for U
        dU = 971761939728640320549601132085879836204587084162
        self.assertEqual(
            format(dU, str(ec.nsize) + "x"), "aa374ffc3ce144e6b073307972cb6d57b2a4e982"
        )
        QU = mult(dU, ec.G, ec)
        self.assertEqual(
            QU,
            (
                466448783855397898016055842232266600516272889280,
                1110706324081757720403272427311003102474457754220,
            ),
        )
        self.assertEqual(
            bytes_from_point(QU, ec).hex(), "0251b4496fecc406ed0e75a24a3c03206251419dc0"
        )

        # 2.1.3 Signing Operation for U
        msg = b"abc"
        k = 702232148019446860144825009548118511996283736794
        exp_sig = (
            0xCE2873E5BE449563391FEB47DDCBA2DC16379191,
            0x3480EC1371A091A464B31CE47DF0CB8AA2D98B54,
        )
        sig = dsa.sign(msg, dU, k, ec, hf)
        r, s = sig
        self.assertEqual(r, exp_sig[0])
        self.assertIn(s, (exp_sig[1], ec.n - exp_sig[1]))

        # 2.1.4 Verifying Operation for V
        self.assertTrue(dsa.verify(msg, QU, sig, ec, hf))

    def test_low_cardinality(self):
        """test low-cardinality curves for all msg/key pairs."""

        # ec.n has to be prime to sign
        prime = [11, 13, 17, 19]

        # only low card or it would take forever
        for ec in low_card_curves.values():
            if ec.p in prime:  # only few curves or it would take too long
                for q in range(1, ec.n):  # all possible private keys
                    PJ = _mult_jac(q, ec.GJ, ec)  # public key
                    for e in range(ec.n):  # all possible int from hash
                        for k in range(1, ec.n):  # all possible ephemeral keys
                            RJ = _mult_jac(k, ec.GJ, ec)
                            Rx = (RJ[0] * mod_inv(RJ[2] * RJ[2], ec.p)) % ec.p
                            r = Rx % ec.n
                            s = mod_inv(k, ec.n) * (e + q * r) % ec.n
                            # bitcoin canonical 'low-s' encoding for ECDSA
                            if s > ec.n / 2:
                                s = ec.n - s
                            if r == 0 or s == 0:
                                self.assertRaises(ValueError, dsa._sign, e, q, k, ec)
                                continue

                            sig = dsa._sign(e, q, k, ec)
                            self.assertEqual((r, s), sig)
                            # valid signature must pass verification
                            self.assertIsNone(dsa._verhlp(e, PJ, r, s, ec))

                            JacobianKeys = dsa._recover_pubkeys(e, r, s, ec)
                            Qs = [ec._aff_from_jac(key) for key in JacobianKeys]
                            self.assertIn(ec._aff_from_jac(PJ), Qs)

    def test_pubkey_recovery(self):
        ec = secp112r2
        q = 0x10
        Q = mult(q, ec.G, ec)
        msg = "Satoshi Nakamoto"
        k = None
        sig = dsa.sign(msg, q, k, ec)
        self.assertTrue(dsa.verify(msg, Q, sig, ec))
        dersig = dsa.serialize(*sig, ec)
        self.assertTrue(dsa.verify(msg, Q, dersig, ec))
        r, s = dsa.deserialize(dersig)
        self.assertEqual((r, s), sig)

        keys = dsa.recover_pubkeys(msg, sig, ec)
        self.assertEqual(len(keys), 4)
        self.assertIn(Q, keys)
        for Q in keys:
            self.assertTrue(dsa.verify(msg, Q, sig, ec))

    def test_crack_prvkey(self):
        ec = secp256k1

        q = 0xDEADBEEF6A307F426A94F8114701E7C8E774E7F9A47E2C2035DB29A206321725
        k = 1010101010101010101

        msg1 = "Paolo is afraid of ephemeral random numbers"
        sig1 = dsa.sign(msg1, q, k)
        # print(f'\nr1: {hex(sig1[0]).upper()}')
        # print(f's1: {hex(sig1[1]).upper()}')

        msg2 = "and Paolo is right to be afraid"
        sig2 = dsa.sign(msg2, q, k)
        # print(f'\nr2: {hex(sig2[0]).upper()}')
        # print(f's2: {hex(sig2[1]).upper()}')

        qc, kc = dsa.crack_prvkey(msg1, sig1, msg2, sig2)
        self.assertEqual(q, qc)
        self.assertIn(k, (kc, ec.n - kc))

        self.assertRaises(ValueError, dsa.crack_prvkey, msg1, sig1, msg2, (16, sig1[1]))
        self.assertRaises(ValueError, dsa.crack_prvkey, msg1, sig1, msg1, sig1)


def test_forge_hash_sig():
    """forging valid hash signatures"""

    ec = secp256k1

    # see https://twitter.com/pwuille/status/1063582706288586752
    # Satoshi's key
    key = "03 11db93e1dcdb8a016b49840f8c53bc1eb68a382e97b1482ecad7b148a6909a5c"
    P = point_from_octets(key, ec)

    # pick u1 and u2 at will
    u1 = 1
    u2 = 2
    R = double_mult(u2, P, u1, ec.G, ec)
    r = R[0] % ec.n
    u2inv = mod_inv(u2, ec.n)
    s = r * u2inv % ec.n
    e = s * u1 % ec.n
    dsa._verhlp(e, (P[0], P[1], 1), r, s, ec)

    # pick u1 and u2 at will
    u1 = 1234567890
    u2 = 987654321
    R = double_mult(u2, P, u1, ec.G, ec)
    r = R[0] % ec.n
    u2inv = mod_inv(u2, ec.n)
    s = r * u2inv % ec.n
    e = s * u1 % ec.n
    dsa._verhlp(e, (P[0], P[1], 1), r, s, ec)


if __name__ == "__main__":
    # execute only if run as a script
    unittest.main()  # pragma: no cover



#!/usr/bin/env python3

# Copyright (C) 2017-2020 The btclib developers
#
# This file is part of btclib. It is subject to the license terms in the
# LICENSE file found in the top-level directory of this distribution.
#
# No part of btclib including this file, may be copied, modified, propagated,
# or distributed except according to the terms contained in the LICENSE file.

""" Hash based helper functions.

"""

from typing import Optional, Tuple

from .alias import Key, Script
from .script import encode
from .to_pubkey import pubkeyinfo_from_key
from .utils import hash160, sha256

_H160Net = Tuple[bytes, str]


def hash160_from_pubkey(
    key: Key, network: Optional[str] = None, compressed: Optional[bool] = None
) -> _H160Net:
    pubkey, network = pubkeyinfo_from_key(key, network, compressed)
    h160 = hash160(pubkey)
    return h160, network


def hash160_from_script(script: Script) -> bytes:
    if isinstance(script, list):
        script = encode(script)
    return hash160(script)


def hash256_from_script(script: Script) -> bytes:
    if isinstance(script, list):
        script = encode(script)
    return sha256(script)
